/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axconn.c
   Applix ASC version: 15
   Date: Wed Feb 24 21:13:21 1999
*/
#include <aos.h>
#include <axhtab.h>
#include <axconn.h>

static axconn *newConn();
static void deleteConn(), deleteLink();
static int newLink();
static int locateRemote(), nextEvent();
static void *iLinkFunc();
static int linkTypeFromIndex();

#if !CENTRAL_EV
static char *newEventNum();
static void tossEventNum();
#endif

static char *newLinkType();
static char *newLocatorFunc();
static char *newLinkFunc();
static char *newLinkEndpoint();
static char *newEventNum();
static void tossLocatorFunc();
static void tossRemoteFunc();
static void tossLinkType();
static void tossLinkEndpoint();
static void tossLinkFunc();
static void tossEventNum();

axconnC *
axNewConnScope(scopeName, namelen)
void *scopeName;
int namelen;
{
	axconnC *scope;
	
	scope = (axconnC *) malloc(sizeof(axconnC));
	scope->name = (void *) malloc(namelen);
	memcpy((void *) scope->name, scopeName, namelen);
	scope->namelen = namelen;
	scope->conn = newConn;
	scope->dconn = deleteConn;
	scope->link = newLink;
	scope->dlink = deleteLink;
	scope->locate = locateRemote;
	scope->linkCall = iLinkFunc;
#if CENTRAL_EV
	scope->newEvent = nextEvent;
#endif
	scope->evnum = 1;
	axhInitTable(&scope->locators, 11, newLocatorFunc, tossLocatorFunc);
	axhInitTable(&scope->linkFuncs, 11, newLinkFunc, tossLinkFunc);
	
	return (scope);
}

void
axDeleteConnScope(scope)
axconnC *scope;
{
	if (scope) {
		if (scope->name) {
			free(scope->name);
		}
		axhDeleteTable(&scope->locators, (void *) NULL);
		axhDeleteTable(&scope->linkFuncs, (void *) NULL);
		free(scope);
	}
}

void
axRegisterLinkFunc(scope, linkType, func,
				   linkDataCopier, linkDataDestructor)
axconnC *scope;
int linkType;
void *(*func)();  /* app-specified func for doing link work */
void *(*linkDataCopier)();    /* app-specified func */
void (*linkDataDestructor)(); /* app-specified func */
{
	char *ctmp;
	link_funcs lf;

	lf.func = func;
	lf.linkDataCopier = linkDataCopier;
	lf.linkDataDestructor = linkDataDestructor;
	axhGetItemI((void *) &linkType, sizeof(int), &scope->linkFuncs,
				(char *) &lf, (char **) &ctmp);
}

void
axRegisterConnLocator(scope, linkType, locatorFunc)
axconnC *scope;
int linkType;
int (*locatorFunc)();
{
	char *ctmp;
	axhGetItemI((void *) &linkType, sizeof(int), &scope->locators,
				(char *) locatorFunc, (char **) &ctmp);
}

void
axDeregisterConnLocator(scope, linkType)
axconnC *scope;
int linkType;
{
	char *ctmp;
	axhTossItemI((void *) &linkType, sizeof(int),
				 &scope->locators, (char *) 0);
}

static void *
iLinkFunc(scope, evnum, from, to, namelen, linkType, data)
axconnC *scope;
int evnum;
void *from;      /* name of calling object */
void *to;        /* name of called object */
int namelen;   /* length in bytes of from, to */
int linkType;
char *data;      /* link-func-specific data */
{
	int ev, *lastev;
	char *localObj;
	axconn **localConn;
	link_funcs *appFuncs;
	char *retVal = data;

	if (!from) {
		/* this is the start of an event
		 */
		ev = scope->newEvent(scope, linkType);    /* central-ev style */
	}
	else {
		ev = evnum;
	}
	if (!scope->locate(scope, linkType, to, &localObj, &localConn))
		return ((char *) 0);

	if (!localConn)
		return ((char *) 0);

	if (from && *localConn) {
		/* get #of last event received from 'from'
		 */
		axhGetItemI((void *) from, namelen, &(*localConn)->evTable,
				   (char *) 0, (char **) &lastev);
		if (ev > *lastev) {
			/* a new event
			 */
			*lastev = ev;                 /* accept new event# */
			
			/* call app linkfunc
			 */
			axhGetItemI((void *) &linkType, sizeof(int),
						&scope->linkFuncs, (char *) 0, (char **) &appFuncs);
			if (appFuncs)
				retVal = appFuncs->func(scope, ev, from, to,
										namelen, linkType, data);
		}
	}
	else {
		/* no connector, or 1st call, just go
		 */
		if (*localConn) {
			(*localConn)->evnum = ev;
		}
		axhGetItemI((void *) &linkType, sizeof(int), 
					&scope->linkFuncs, (char *) 0, (char **) &appFuncs);
		if (appFuncs)
			retVal = appFuncs->func(scope, ev, from, to,
									namelen, linkType, data);
	}
	return retVal;
}

static axconn *
newConn()
{
	axconn *tconn;

	tconn = (axconn *) malloc(sizeof(axconn));
	memset((char *) tconn, 0, sizeof(axconn));
	axhInitTable(&tconn->typeTable, 5, newLinkType, tossLinkType);

	/* initialize this' evnum, evtable
	 */
	tconn->evnum = 0;
	axhInitTable(&tconn->evTable, 11, newEventNum, tossEventNum);
	
#if !CENTRAL_EV
	/* init this' id, event number
	 */
#endif	
	
	return (tconn);
}

static void
deleteConn(scope, name, namelen, tconn)
axconnC *scope;
void *name;      /* name of containing object */
int namelen;     /* length of name in bytes */
axconn *tconn;   /* object's conn handle */
{
	int ix, jx;
	axhtab *tabl;
	hitem *list;
	int linkType;

	if (tconn) {
		/* must destruct links tables first
		 */
		for (ix = 0; ix < tconn->numLinkTypes; ix++) {
			tabl = &tconn->linkTables[ix];
			linkType = linkTypeFromIndex(tconn, ix);
			/* use the (reciprocal) deleteLink here
			 */
			for (jx = 0; jx < tabl->width; jx++) {
				list = tabl->table[jx];
				while(list) {
					if (list->value && list->name) {
						deleteLink(scope, linkType, name, list->name,
								   namelen, TRUE);
						list->value = (char *) NULL;
					}
					list = list->next;
				}
			}
			axhDeleteTable(&tconn->linkTables[ix], (void *) NULL);
		}
		free((char *) tconn->linkTables);

		axhDeleteTable(&tconn->typeTable, (void *) NULL);
		axhDeleteTable(&tconn->evTable, (void *) NULL);
		free((char *) tconn);
	}
}

static int
locateRemote(scope, linkType, name, remoteObj, remoteConn)
axconnC *scope;        /* in */
int linkType;          /* in */
void *name;            /* in: name of remote obj */
char **remoteObj;      /* out */
axconn ***remoteConn;  /* out */
{
	remoteRef ref;
	int (*locator)();

	memset((char *) &ref, 0, sizeof(remoteRef));
	axhGetItemI((void *) &linkType, sizeof(int), &scope->locators,
				(char *) 0, (char **) &locator);
	if (locator) {
		locator(scope, name, &ref);
	}
	if (ref.rHandle && ref.rConn) {
		*remoteObj = ref.rHandle;   /* address of remote obj */
		*remoteConn = ref.rConn;    /* *address of its axconn */
		return TRUE;
	}
	else {
		*remoteObj = (char *) 0;
		*remoteConn = (axconn **) 0;
		return FALSE;
	}
}

static int
newLink(scope, linkType, from, to, namelen, data, doBack)
axconnC *scope;
int linkType;
void *from;    /* name of local object */
void *to;      /* name of remote object */
int namelen;   /* length of to,from names in bytes */
char *data;    /* data re: the link */
int  doBack;   /* create a back-link also */
{
	axconn **localConn, *lcon;
	char *localObj, *ctmp;
	axlink *tlink;
	int typeIndex;
	axhtab *linkTab;

	/* find myself
	 */
	if (!scope->locate(scope, linkType, from, &localObj, &localConn))
		return (FALSE);

	if (!localConn)
		return (FALSE);

	if (!(*localConn)) {
		/* no local conn: create
		 */
		if (!(*localConn = scope->conn()))
			return (FALSE);
	}

	if (doBack) {
		/* add a link from remote to self
		 */
		if (!scope->link(scope, linkType, to, from, namelen, data, FALSE)) {
			return (FALSE);
		}
	}

	/* add new link endpoint to local conn
	 */

	/* create the endpoint
	 */
	tlink = (axlink *) malloc(sizeof(axlink));
	tlink->scope = scope;
	tlink->linkType = linkType;
	tlink->direction = doBack;
	if (data) {
		link_funcs *lf;
		/* look for a copier for the data
		 */
		axhGetItemI((void *) &linkType, sizeof(int), &scope->linkFuncs,
				   (char *) NULL, (char **) &lf);
		if (lf->linkDataCopier) {
			lf->linkDataCopier((char *) data,
							   (char **) &tlink->data);
		}
		else {
			/* no copier; assume integer's worth of freestore data
			 */
			tlink->data = (char *) malloc(sizeof(int));
			memcpy((char *) tlink->data, data, sizeof(int));
		}
	}

	/* get the index for the link type
	 */
	lcon = *localConn;
	axhGetItemI((void *) &linkType, sizeof(int), &lcon->typeTable,
				(char *) lcon, (char **) &typeIndex);

	/* load the link endpoint into its table
	 */
	linkTab = &lcon->linkTables[typeIndex];
	if (!linkTab->width) {
		/* a new type for this conn; init new table
		 */
		memset((char *) linkTab, 0, sizeof(axhtab));
		axhInitTable(linkTab, 11, newLinkEndpoint, tossLinkEndpoint);
	}
	axhGetItemI((void *) to, namelen, linkTab,
				(char *) tlink, (char **) &ctmp);
	tossLinkEndpoint((char *) tlink);
	return (TRUE);
}

static void
deleteLink(scope, linkType, from, to, namelen, doBack)
axconnC *scope;
int linkType;
void *from;       /* name of local object */
void *to;         /* name of remote object */
int namelen;      /* length of to,from names in bytes */
int  doBack;      /* delete the back-link also */
{
	char *localObj;
	axconn **localConn, *lcon;
	int typeIndex;
	
	/* find myself
	 */
	if (!scope->locate(scope, linkType, from, &localObj, &localConn))
		return;

	if (doBack) {
		/* remove the remote link to me
		 */
		scope->dlink(scope, linkType, to, from, namelen, FALSE);
	}

	/* remove the local endpoint
	 */
	lcon = *localConn;
	axhGetItemI((void *) &linkType, sizeof(int), &lcon->typeTable,
				(char *) 0, (char **) &typeIndex);
	axhTossItemI((void *) to, namelen,
				 &lcon->linkTables[typeIndex], (char *)0);
}

#if CENTRAL_EV
static int
nextEvent(scope, linkType)
axconnC *scope;
int linkType;
{
	if (scope->evnum == MAXEVENT) {
		/* event overflow;  reset everyone's evnum..
		 */
		int ix, jx;
		hitem *list, *elist;
		char *owner;
		axconn **tconn, *con;

		scope->evnum = 0;
		for (ix = 0; ix < scope->locators.width; ix++) {
			list = scope->locators.table[ix];
			while (list) {
				if (scope->locate(scope, linkType, list->name,
								  &owner, &tconn)) {
					
					if (tconn && (con = *tconn)) {
						for (jx = 0; jx < con->evTable.width; jx++) {
							elist = con->evTable.table[jx];
							while(elist) {
								/* the event number is stored as
								 * a free-store int
								 */
								*elist->value = 0; /* reset */
							}
							elist = elist->next;
						}
					}
				}
				list = list->next;
			}
		}
	}
	scope->evnum++;
	return(scope->evnum);
}
#else
/* note that for distributed event management, it is necessary for
   the local conn to maintain a table of all remotes from which it
   has ever received "interesting" messages.  "interesting" means
   messages other than wraparound-notify and destruct-notify.
   the local conn must drop wrap-notify if it has never received
   interesting messages from that remote.  destruct-notify always
   implies the removal of that remote's table item, if it was ever
   there.  both wrap-notify and destruct-notify must propagate as
   fully as possible. */
#endif

/*
 * Constructors/destructors for items in the conn
 * class, and in conn instances
 */

/* a link type item value is an integer representing
 * which element of a conn instance's linkTables array
 * contains the appropriate links table
 */
static char *
newLinkType(typename, notUsed, connAddr)
void *typename;
int notUsed;
void *connAddr;
{
	axconn *thisconn = (axconn *) connAddr;
	if (!thisconn->linkTables) {
		thisconn->linkTables = (axhtab *) malloc(sizeof(axhtab));
		memset((char *) thisconn->linkTables, 0, sizeof(axhtab));
		thisconn->numLinkTypes = 1;
	}
	else {
		thisconn->linkTables = (axhtab *) realloc
			(thisconn->linkTables,
			 (thisconn->numLinkTypes + 1) * sizeof(axhtab));
		memset((char *) &thisconn->linkTables[thisconn->numLinkTypes],
			   0, sizeof(axhtab));
		thisconn->numLinkTypes++;
	}
	return ((char *) thisconn->numLinkTypes - 1);
}

static void
tossLinkType(notUsed, notUsed2)
char *notUsed;
char *notUsed2;
{
	return;
}

#if CENTRAL_EV
/* an event num item value is an integer representing
 * the last event number received from 'fromlink'
 */
static char *
newEventNum(fromlink, notUsed, notUsed2)
void *fromlink;
int notUsed;
void *notUsed2;
{
	/* though just an integer, allocate this so
	 * that it's easy to update during operation
	 */
	int *eventNumber = (int *) malloc(sizeof(int));
	
	*eventNumber = 0;
	return ((char *) eventNumber);
}

static void
tossEventNum(eventNumberP, notUsed2)
void *eventNumberP;
void *notUsed2;
{
	if (eventNumberP)
		free((char *) eventNumberP);
}
#else
    /* an event num item value is a hash table associating
	 * remote object id's with the last event number received
	 * from them along fromlink
	 */
#endif


/* a locator func item value is a pointer to a
 * function which returns the address of the
 * named object, and the address of its conn handle
 */
static char *
newLocatorFunc(name, notUsed, funcptr)
void *name;        /* name of object */
int notUsed;
void *funcptr;     /* interpret as int (*)() */
{
	return ((char *) funcptr);
}

static void
tossLocatorFunc(ref, notUsed)
void *ref;
void *notUsed;
{
	return;
}

/* a link func item value is a link_funcs
 * instance, which holds a trio of app-specified
 * function pointers: one for doing 'link work',
 * one for copying link data, and one for destroying
 * link data
 */
static char *
newLinkFunc(notUsed, notUsed2, linkFuncs)
void *notUsed;
int notUsed2;
void *linkFuncs;
{
	link_funcs *lf = (link_funcs *) linkFuncs;
	link_funcs *newFuncs = (link_funcs *) NULL;
	if (lf) {
		newFuncs = (link_funcs *) malloc (sizeof(link_funcs));
		*newFuncs = *lf;
	}
	return((char *) newFuncs);
}

static void
tossLinkFunc(funcNames, notUsed)
void *funcNames;   /* (link_funcs *) */
void *notUsed;
{
	if (funcNames) {
		free((char *) funcNames);
	}
}

static char *
newLinkEndpoint(name, notUsed, linkt)
void *name;
int notUsed;    /* arrives as name length in bytes */
void *linkt;    /* a pointer to a axlink_t */
{
	axlink *ll = (axlink *) linkt;
	axlink *newLink = (axlink *) malloc (sizeof(axlink));

	*newLink = *ll;
	
	if (ll->data) {
		link_funcs *lf;
		/* copy in the link data
		 */
		        /* first see whether there's a copy func
				 */
		axhGetItemI((void *) &ll->linkType, sizeof(int), &ll->scope->linkFuncs,
				   (char *) NULL, (char **) &lf);
		if (lf->linkDataCopier) {
			lf->linkDataCopier((char *) ll->data,
							   (char **) &newLink->data);
		}
		else {
			    /* no copier; assume integer's worth of freestore data
				 */
			newLink->data = (char *) malloc(sizeof(int));
			memcpy((char *) newLink->data, ll->data, sizeof(int));
		}
	}
	
	return((char *) newLink);
}

static void
tossLinkEndpoint(linkt, notUsed)
void *linkt;
void *notUsed;
{
	axlink *ll = (axlink *) linkt;
	if (ll) {
		if (ll->data) {
			link_funcs *lf;
			/* free the link data
			 */
			        /* first see whether there's a destructor func
					 */
			axhGetItemI((void *) &ll->linkType, sizeof(int),
						&ll->scope->linkFuncs, (char *) NULL, (char **) &lf);
			if (lf->linkDataDestructor) {
				lf->linkDataDestructor((char *) ll->data);
			}
			else {
				    /* no destructor; assume single block freestore mem
					 */
				free((char *) ll->data);
			}
		}
		free((char *) ll);
	}
}

static int
linkTypeFromIndex(cp, ix)
axconn *cp;
int ix;
{
	int jx, linkType;
	hitem *list;
	axhtab *tabl = &cp->typeTable;
	
	for (jx = 0; jx < tabl->width; jx++) {
		list = tabl->table[jx];
		while(list) {
			if (list->name && ((long) list->value) == ix) {
				memcpy((void *) &linkType, (void *) list->name, sizeof(int));
				return (linkType);
			}
			list = list->next;
		}
	}
	return(0);
}



