/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axgfscfg.c
   Applix ASC version: 9
   Date: Mon Nov 16 13:53:36 1998
*/
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <errno.h>
#include <aos.h>
#include <axtext.h>
#include <axhtab.h>
#include <axgfs.h>
#include <axtrutyp.h>

#define MAX_XLFD 257

#define WHITESPACE(c) (c == ' ' || c == '\t' || c == '\n' || \
                       c == COMMACHAR || c == EQUALCHAR || \
                       c == COMMENTCHAR)

#define COMMENTCHAR ('#')
#define ESCCHAR ('\\')
#define QUOTECHAR ('"')
#define COMMACHAR (',')
#define EQUALCHAR ('=')
#define NEWLINECHAR ('\n')

/* these filenames are those of the encoding files we ship
 */
#define ADOBE_ENC    "tmr.afm"     /* this afm is in AdobeStandardEncoding */
#define LATIN1_ENC   "latin1.enc"
#define BICS_ENC     "bics.enc"
#define UNSPECIFIED  "unspecified"  /* not a file, just a token */

/* most popular fields of the exciting XLFD name format
 */
#define FAMILY_NAME       2
#define WEIGHT_NAME       3
#define SLANT             4
#define SETWIDTH_NAME     5
#define CHARSET_REGISTRY 13
#define CHARSET_ENCODING 14

typedef struct catalogue_t {
	int numDirs;
	char **dirs;
	axhtab *screenNames;
} catalogue;

typedef struct xlfd_style_t {
	char *xlfd;               /* just a reference, do not free */
	int weight;
	int slant;
} xlfd_style;

typedef struct xfamily_t {
	int numXlfds;
	xlfd_style *xlfds;
} xfamily;

typedef struct default_encodings_t {
	char *truetype;
	char *speedo;
	char *type1;
} default_encodings;

/* this "parse state" typedef is used for reading a fontserver config file
 */
typedef struct pstate_t {
    unsigned int intext : 1;
    unsigned int incomment : 1;
    unsigned int inquotes : 1;
	unsigned int incatalogue : 1;
    char lastchar;
    char *lastXblack;
    char *filedata;
	catalogue cg;
} pstate;

extern void axtFormScalableName();
extern void loadFontRecord();
extern void freeErrString();
extern int axRealDisplayResolution();
extern int axttFontEncoding();
extern char *axGetFontRecord();

char *axgfsReadFonts();

static void evalTok();
static void deleteFontfileXLFD();
static void deleteFam();
static void readFamilyName();
static void capWords();
static void measureStyle();
static void readField();
static void splitFams();
static void formFamilyName();
static void readEncodingAndMetricsNames();
static int isBlack();
static int isPathname();
static char *printError();
static char *readFontNames();
static char *readConfig();
static char *strippedCopy();
static char *newFontfileXLFD();
static char *newFam();
static char *evalPath();

/* : axgfsReadFonts - read the Gallium font server's config
 * file, and load a representative set of font records.
 * A NULL return value indicates no error.  A nonnull return
 * is a charstring describing some error condition, caller must free.
 */
char *                                   /* caller must free nonnull return */
axgfsReadFonts(configPath, screenNames)
char *configPath;     /* in: config file pathname */
axhtab *screenNames;  /* in/out: update with gfs screenfont info */
{
	pstate ps;
	char *dirname, *err, *terr;
	char tName[MAX_XLFD];
	int ix, jx, kx, itmp, numFams;
	axhtab fams;
	axhtab finalFams;
	xfamily *tfam, **ffams;
	xlfd_style *xs;
	fontfile_xlfd *xf;
	hitem *list;
	fontTabEnt ftab, *fcacheRef;
	default_encodings encFiles;

	err = (char *) 0;
	dirname = (char *) 0;
	memset((void *) screenNames, 0, sizeof(axhtab));
	memset((void *) &ps, 0, sizeof(pstate));
	memset((void *) &fams, 0, sizeof(axhtab));
	memset((void *) &finalFams, 0, sizeof(axhtab));
	
	/* assign members of the default encodings struct
	 */
	encFiles.truetype = LATIN1_ENC;
	encFiles.speedo = BICS_ENC;
	encFiles.type1 = ADOBE_ENC;
	
	/* read config file to collect a list of
	 * catalogue directories.
	 */
	ps.cg.screenNames = screenNames;
	err = readConfig(&ps, configPath);
	if (err) {
		goto free_resources;
	}
	
	/* for each directory, open the fonts.dir file if present,
	 * and store an array of xlfd/fontfile-name pairings
	 */
	for (ix = 0, dirname = (char *) NULL; ix < ps.cg.numDirs; ix++) {
		itmp = strlen(ps.cg.dirs[ix]) + 1;
		if (dirname) {
			dirname = (char *) realloc(dirname, itmp);
		}
		else {
			dirname = (char *) malloc(itmp);
		}
		strcpy(dirname, ps.cg.dirs[ix]);
		err = readFontNames(&ps.cg, dirname, configPath);
		if (err) {
			goto free_resources;
		}
	}
	
	/* create an array of font families; each element is itself
	 * an array of those above xlfds that fall in each family
	 */
	axhInitTable(&fams, 31, newFam, deleteFam);
	for (ix = 0; ix < ps.cg.screenNames->width; ix++) {
		list = ps.cg.screenNames->table[ix];
		while(list) {
			/* get the font family name from the xlfd
			 */
			readFamilyName(list->name, tName);
			
			/* record this family name
			 */
			axhGetItem(tName, &fams, (char *) 0, (char **) &tfam);
			
			/* append the xlfd reference onto this fam's list of xlfds.
			 */
			if (tfam->xlfds) {
				tfam->xlfds = (xlfd_style *) realloc
					(tfam->xlfds, (tfam->numXlfds + 1) * sizeof(xlfd_style));
			}
			else {
				tfam->xlfds = (xlfd_style *) malloc
					((tfam->numXlfds + 1) * sizeof(xlfd_style));
			}
			xs = &tfam->xlfds[tfam->numXlfds];
			xs->xlfd = list->name;
			measureStyle(xs);    /* make wild guesses as to weight, slant */
			tfam->numXlfds++;
			
			list = list->next;
		}
	}
	
	/* go thru the fams table, and split the first-cut family
	 * groupings onto finalFams.  the final families aren't allowed
	 * to have more than 4 members, because of our apps' assumptions
	 * that a family's fonts are fully selectable using boolean bold/ital
	 */
	axhInitTable(&finalFams, 31, newFam, deleteFam);
	for (ix = 0; ix < fams.width; ix++) {
		list = fams.table[ix];
		while(list) {
			tfam = (xfamily *) list->value;
			splitFams(list->name, tfam, &finalFams);
			list = list->next;
		}
	}

	/* spin thru final fams, generate and load fontTabEnt instances.
	 */
	for (ix = 0; ix < finalFams.width; ix++) {
		list = finalFams.table[ix];
		while(list) {
			tfam = (xfamily *) list->value;

			for (jx = 0; jx < tfam->numXlfds; jx++) {
				memset((void *) &ftab, 0, sizeof(fontTabEnt));
				
				ftab.xNamesKnown = TRUE;
				ftab.scalable = TRUE;
				ftab.gallium = TRUE;
				ftab.use.xNamesKnown = ftab.use.scalable =
					ftab.use.gallium = TRUE;
				
				/* global name must be unique
				 */
				formFamilyName(tfam->xlfds[jx].xlfd, FALSE, tName);
				ftab.globalName = tName;
				ftab.use.globalName = TRUE;
				
				ftab.familyName = list->name;
				ftab.use.familyName = TRUE;
				
				/* weight and slant are assigned as booleans;
				 * a guess is used for weight, while for slant
				 * the bool val is taken directly from the xfamily instance.
				 */
				if (tfam->numXlfds > 1) {
					/* assign weight bool based on the high/low vals
					 */
					int minv, maxv;
				    maxv = -1;
					minv = INT_MAX;
					for (kx = 0; kx < tfam->numXlfds; kx++) {
						if (tfam->xlfds[kx].weight < minv)
							minv = tfam->xlfds[kx].weight;
						if (tfam->xlfds[kx].weight > maxv)
							maxv = tfam->xlfds[kx].weight;
					}
					ftab.weight = tfam->xlfds[jx].weight == maxv;
				}
				else {
					ftab.weight = tfam->xlfds[jx].weight > 2;
				}
				ftab.slant = tfam->xlfds[jx].slant;
				ftab.use.weight = ftab.use.slant = TRUE;

				ftab.xNameCount = 1;
				ftab.xNames = (char **) malloc(sizeof(char *));
				ftab.xNames[0] = tfam->xlfds[jx].xlfd;
				ftab.use.xNames = ftab.use.xNameCount = TRUE;
				
				/* encodingName and metricsName are derived from
				 * the font file name
				 */
				readEncodingAndMetricsNames(tfam->xlfds[jx].xlfd, 
											ps.cg.screenNames, &encFiles,
											&ftab, &xf);
				
				/* load the new record
				 */
				loadFontRecord(&ftab);
				
				/* lookup the fontcache address of the newly loaded
				 * record, and record this ref on the fontfile_xlfd
				 * instance associated with tfam->xlfds[jx].xlfd
				 */
				terr = axGetFontRecord(ftab.globalName, &fcacheRef);
				if (terr) {
					freeErrString(terr);
					xf->ftab = (fontTabEnt *) NULL;
				}
				else {
					xf->ftab = fcacheRef;
				}

				/* free those ftab members allocated during this block
				 */
				if (ftab.numTargetMetricsNames) {
					for (kx = 0; kx < ftab.numTargetMetricsNames; kx++) {
						free(ftab.targetMetricsNames[kx]);
						free(ftab.targetMetricsNames);
					}
				}
				if (ftab.xNames) {
					free(ftab.xNames);
				}
			}
			
			list = list->next;
		}
	}

 free_resources:
	/* free resources
	 */
	for (ix = 0; ix < ps.cg.numDirs; ix++) {
		if (ps.cg.dirs[ix])
			free(ps.cg.dirs[ix]);
	}
	if (ps.cg.dirs) {
		free(ps.cg.dirs);
	}
	if (fams.width) {
		axhDeleteTable(&fams, (void *) NULL);
	}
	if (finalFams.width) {
		axhDeleteTable(&finalFams, (void *) NULL);
	}
	if (ps.filedata) {
		free(ps.filedata);
	}
	if (dirname) {
		free(dirname);
	}
	return (err);
}

static char *
readConfig(ps, filename)
pstate *ps;
char *filename;
{
	FILE *fp;
	char *cp;
	long filesize;
	
	fp = fopen(filename, "r");
	if (!fp) {
		/* can't even find the config file; return a token
		 * string that axgfsInit will test for.
		 */
		char *tokerr = (char *) malloc(strlen(AXGFS_NO_CONFIG) + 1);
		strcpy(tokerr, AXGFS_NO_CONFIG);
		return ((char *) tokerr);
	}

	/* read config file into memory
	 */
    fseek(fp, 0, 2);
    filesize = ftell(fp);
    ps->filedata = (char *) malloc(filesize + 1);
    fseek(fp, 0, 0);
    fread(ps->filedata, sizeof(char), (int)filesize, fp);
    ps->filedata[filesize] = '\0';
	fclose(fp);
	
	/* read thru the contents to find the catalogue directories
	 */
	cp = ps->filedata;
    while(*cp != '\0') {
        if (isBlack(ps, *cp)) {
            if (!ps->intext) {
				/* transition into blackspace
				 */
                ps->intext = TRUE;
                ps->lastXblack = cp;
            }
        }
        else {
            if (ps->intext) {
				/* transition into whitespace
				 */
				evalTok(ps, cp);
                ps->intext = FALSE;
            }
        }
        
        if (*cp == QUOTECHAR && ps->lastchar != ESCCHAR) {
            ps->inquotes = !ps->inquotes;

        }
        
        ps->lastchar = *cp;
        cp++;
	}
	return ((char *) 0);
}

static char *
readFontNames(cg, dirname, configName)
catalogue *cg;
char *dirname;
char *configName;
{
	FILE *fp;
	char *cp, *ctmp, *fname, *xlfd, curline[2048];
	char *fullname, *err;
	int numFonts;
	fontfile_xlfd xf;

	/* form fullname from dirname, filename
	 */
	fullname = (char *) malloc(strlen(dirname) + 11);
	strcpy(fullname, dirname);
	strcat(fullname, "/fonts.dir");  /* the extra 11 chars */
	fp = fopen(fullname, "r");
	if (!fp) {
#if !FOR_FILTER		
		err = printError
			(XLT2("axgfscfg.c-1", "I can't open font index file \'fonts.dir\',\n"),
			 XLT2("axgfscfg.c-2", "which I expected to find in directory:\n  \'"), dirname,
			 XLT2("axgfscfg.c-3", "\'\nbased on the catalogues list in config file:\n"),
			 configName);
		return (err);
#else
		return ((char *) NULL);
#endif
	}
	
	/* 1st line declares number of fonts in index
	 */
	if (fgets(curline, 2048, fp) == (char *) NULL) {
#if !FOR_FILTER		
		err = printError
			(XLT2("axgfscfg.c-4", "I can't understand the format of font index file \'fonts.dir\',\n"),
			 XLT2("axgfscfg.c-5", "which I found in directory:\n  \'"), dirname,
			 XLT2("axgfscfg.c-6", "\'\nbased on the catalogues list in config file:\n"),
			 configName);
		free(fullname);
		return (err);
#else
		return ((char *) NULL);
#endif
	}
	else {
		if(!sscanf(curline, "%d", &numFonts)) {
#if !FOR_FILTER			
			err = printError
				(XLT2("axgfscfg.c-4", "I can't understand the format of font index file \'fonts.dir\',\n"),
				 XLT2("axgfscfg.c-5", "which I found in directory:\n  \'"), dirname,
				 XLT2("axgfscfg.c-7", "\'\nbased on the catalogues list in config file:\n"),
				 configName);
			free(fullname);
			return (err);
#else
			return ((char *) NULL);
#endif
		}
	}
	while (numFonts && (fgets(curline, 2048, fp) != (char *) NULL)) {
		/* This fonts.dir line parsing code is real quicknnasty. Assumptions:
		 * - filenames always start at column 0
		 * - filenames always have a ".<xyz>" suffix, e.g., ".ttf"
		 * - xlfd names are always well-formed
		 */
		fname = cp = curline;
		
		/* advance to beginning of suffix
		 */
		while (*cp && *cp != '.') {
			cp++;
		}
		
		/* advance to just past suffix, null-stamp, skip null
		 */
		cp += 4;
		*cp = '\0';
		cp++;
		
		/* advance to 1st dash of xlfd name
		 */
		while (*cp && *cp != '-') {
			cp++;
		}
		xlfd = cp;

		/* advance to just past xlfd, null-stamp
		 */
		while (*cp && *cp != '\n') {
			cp++;
		}
		*cp = '\0';
		
		/* backup over any consecutive whitespace chars
		 */
		cp--;
		while (WHITESPACE(*cp)) {
			*cp = '\0';
			cp--;
		}
		
		/* store filename-xlfd pair
		 */
		if (!cg->screenNames->width) {
			axhInitTable(cg->screenNames, 31,
						 newFontfileXLFD, deleteFontfileXLFD);
		}
		xf.xlfd_template = xlfd;
		xf.dir = dirname;
		xf.filename = fname;
		axhGetItem(xlfd, cg->screenNames, (char *) &xf, &ctmp);
		
		numFonts--;
	}

	fclose(fp);
	free(fullname);
	return ((char *) 0);
}

static int
isBlack(p, ch)
pstate *p;
char ch;
{
    if (WHITESPACE(ch)) {
        if (!p->incomment) {
            if ( (ch == COMMENTCHAR) && 
                 (p->lastchar != ESCCHAR) ) {
                p->incomment = TRUE;
            }
        }
        else if (p->incomment) {
            if (ch == NEWLINECHAR && p->lastchar != ESCCHAR) {
                p->incomment = FALSE;
            }
        }
        if (p->intext) {
            return (p->lastchar == ESCCHAR || p->inquotes);
        }
        return FALSE;
    }
    else {
        return (!p->incomment);
    }
}

static void
evalTok(ps, cp)
pstate *ps;
char *cp;
{
	char *tok, *expandedPath;
	
	tok = strippedCopy(ps, cp);
	if (tok) {
		if (ps->incatalogue) {
			if (isPathname(tok)) {
				/* looks like a pathname; take a copy that
				 * expands any environment var terms.
				 */
				expandedPath = evalPath(tok);
				
				/* add to list of catalogue dirs
				 */
				if (ps->cg.dirs) {
					ps->cg.dirs = (char **) realloc
						(ps->cg.dirs, (ps->cg.numDirs + 1) * sizeof(char *));
				}
				else {
					ps->cg.dirs = (char **) malloc
						((ps->cg.numDirs + 1) * sizeof(char *));
				}
				ps->cg.dirs[ps->cg.numDirs] = expandedPath;
				(ps->cg.numDirs)++;
				
				return;
			}
			else {
				ps->incatalogue = FALSE;
			}
		}
		else {
			if (!strncmp(tok, "catalogue", 9)) {
				ps->incatalogue = TRUE;
			}
		}
		free(tok);
	}
}

/* : evalPath - param 'inpath' is expected to be an absolute
 * pathname that may contain $VAR terms, where VAR is an environment
 * variable.  Return a copy of inpath with any such terms evaluated.
 * Caller must free returned string.
 */
static char *
evalPath(inpath)
char *inpath;
{
	char *ip, *op, *tcp, *envVal;
	char *expandedPath;
	int olen, env_len, exp_env_len;

	if (!inpath || !strlen(inpath)) {
		return ((char *) NULL);
	}
	
	/* determine size of fully evaluated path; safe estimate is
	 * input path length, plus lengths of all evaluated env var terms.
	 */
	envVal = (char *) NULL;
	for (ip = inpath, olen = strlen(inpath) + 1; *ip; ip++) {
		if (*ip == '$') {
			/* this is taken as the start of an env var expression;
			 * we take its end as just before the next slash char
			 */
			if (*(ip + 1)) {
				for (tcp = ip + 1; *tcp && *tcp != '/'; tcp++) ;

				/* env_len is the string length of the unexpanded
				 * environment variable expression, without the '$'
				 */
				env_len = tcp - ip - 1;

				if (envVal) {
					envVal = (char *) realloc(envVal, env_len + 1);
				}
				else {
					envVal = (char *) malloc(env_len + 1);
				}
				strncpy(envVal, ip + 1, env_len);
				envVal[env_len] = '\0';
				if (tcp = getenv(envVal)) {
					olen += strlen(tcp);
				}
			}
		}
	}
	
	expandedPath = (char *) malloc(olen);
	memset((void *) expandedPath, 0, olen);

	/* repeat scan, this time writing out the expanded path
	 */
	for (ip = inpath, op = expandedPath; *ip; ) {
		if (*ip == '$') {
			for (tcp = ip; *tcp && *tcp != '/'; tcp++) ;
			env_len = tcp - ip - 1;
			if (env_len) {
				if (envVal) {
					envVal = (char *) realloc(envVal, env_len + 1);
				}
				else {
					envVal = (char *) malloc(env_len + 1);
				}
				strncpy(envVal, ip + 1, env_len);
				envVal[env_len] = '\0';
				if (tcp = getenv(envVal)) {
					exp_env_len = strlen(tcp);
					if (exp_env_len) {
						strncpy(op, tcp, exp_env_len);
						op += exp_env_len;
					}
				}
				ip += env_len + 1;
			}
			else {
				*op++ = *ip++;
			}
		}
		else {
			*op++ = *ip++;
		}
	}
	if (envVal) {
		free(envVal);
	}
	return (expandedPath);
}

/* : strippedCopy - make a copy of the string data between
 * p->lastXblack (inclusive) and cp, stripping out escape
 * chars, unescaped quotes, etc.
 */
static char *    /* caller must free nonnull return */
strippedCopy(p, cp)
pstate *p;
char *cp;
{
    char *scopy, *copyp, *tcp, lastchar;
    int len = cp - p->lastXblack;

    if (len <= 0)
        return ((char *) NULL);

	/* copy the input using these rules:
	 *  - QUOTECHARs and ESCCHARs are stripped unless escaped.
	 *  - newlines are stripped unconditionally..
	 *  - nothing is expanded; memory alloc assumes this
	 */
    scopy = (char *) malloc (len + 1);
    memset((char *) scopy, 0, len + 1);
    lastchar = '\0';
    for (tcp = p->lastXblack, copyp = scopy; tcp < cp; tcp++) {
        if ((*tcp == QUOTECHAR || *tcp == ESCCHAR) &&
            lastchar != ESCCHAR) {
            lastchar = *tcp;
            continue;
        }
        if (*tcp == '\n') {
            lastchar = *tcp;
            continue;
        }
        lastchar = *tcp;
        *copyp = *tcp;
        copyp++;
    }
    return scopy;
}

/* : isPathname -- guess whether passed token is a
 * filesystem pathname.  This func is called from
 * the 'incatalogue' state of the config file parser,
 * so absolute pathnames are the expected values.
 */
static int
isPathname(tok)
char *tok;
{
	if (tok && tok[0] == '/') {
		return (TRUE);
	}
	else if (tok && tok[0] == '$') {
		return (TRUE);    /* consider this an environment var expression */
	}
	else {
		return (FALSE);
	}
}

/* : readFamilyName - copy out the FAMILY_NAME field of the passed xlfd.
 * Further qualify the name by concatenating several other xlfd fields.
 * Assume well-formed xlfd.
 */
static void
readFamilyName(xlfd, famName)
char *xlfd;                   /* in */
char *famName;                /* out: onto caller's MAXLFD-sized memory */
{
	int dashcount, itmp;
	char *cp, *tcp, *fcp;
	
	famName[0] = 0;
	for (cp = xlfd, dashcount = 0; *cp; cp++) {
		if (*cp == '-') {
			dashcount++;
		}
		if (dashcount == FAMILY_NAME) {
			/* now at dash just before FAMILY_NAME field
			 */
			cp++;
			tcp = cp;
			while(*tcp != '-') {
				tcp++;
			}
			strncpy(famName, cp, tcp-cp);
			famName[tcp-cp] = '\0';
			break;
		}
	}
	if (!famName[0]) {
		return;
	}
	itmp = strlen(famName);
	
	/* It's necessary to further qualify the family name,
	 * so that if any groups form, they won't be mixed in terms
	 * of several additional characteristics: SETWIDTH_NAME,
	 * CHARSET_REGISTRY, and CHARSET_ENCODING.
	 * (We're trying to form groups selectable using the
	 * application bold/ital settings).
	 *
	 * Rescan the xlfd, and concat these fields using the dash as delimiter.
	 */
	itmp = strlen(famName);
	fcp = &famName[itmp];
	for (cp = xlfd, dashcount = 0; *cp; cp++) {
		if (*cp == '-') {
			dashcount++;
		}
		if (dashcount == SETWIDTH_NAME ||
			dashcount == CHARSET_REGISTRY ||
			dashcount == CHARSET_ENCODING) {

			/* append this field onto famName
			 */
			tcp = cp;
			tcp++;
			while (*tcp && *tcp != '-') {
				tcp++;
			}
			if (tcp > cp) {
				strncpy(fcp, cp, tcp-cp);
				fcp++;   /* point just past last dash on famName */
				fcp += tcp-cp-1;
				if (dashcount == CHARSET_ENCODING) {
					*fcp = '\0';
					break;
				}
				cp += tcp-cp-1;
			}
		}
	}
}

/* : this just returns a particular xlfd field value.
 */
static void
readField(xlfd, fieldNum, fout)
char *xlfd;
int fieldNum;
char *fout;                    /* out: onto caller's MAX_XLFD-sized memory */
{
	char *cp, *tcp;
	int dashcount;
	
	for (cp = xlfd, dashcount = 0; *cp; cp++) {
		if (*cp == '-') {
			dashcount++;
		}
		if (dashcount == fieldNum) {
			cp++;
			tcp = cp;
			while(*tcp && *tcp != '-') {
				tcp++;
			}
			strncpy(fout, cp, tcp-cp);
			fout[tcp-cp] = '\0';
			return;
		}
	}
}

static void
capWords(text, len)
char *text;
int len;
{
	int ix;
	char lastchar;
	
	if (!len) {
		return;
	}
	/* upper-case the 1st char, and whichever following
	 * chars follow whitespace.
	 */
	text[0] = (char) Atoupper((unsigned int) text[0]);
	for (ix = 0, lastchar = 0; ix < len; ix++) {
		if (WHITESPACE(lastchar) && !WHITESPACE(text[ix])) {
			text[ix] = (char) Atoupper((unsigned int) text[ix]);
		}
		lastchar = text[ix];
	}
}

/* : measureStyle - assign numerical values corresponding
 * to the values in the WEIGHT_NAME and SLANT xlfd fields....
 */
static void
measureStyle(xs)
xlfd_style *xs;     /* in/out: read xlfd member, set weight & slant */
{
	char field[MAX_XLFD];
	
	/* rate the slant value; numerical vals are 1 or 0
	 */
	field[0] = (char) 0;
	xs->slant = 0;
	readField(xs->xlfd, SLANT, field);
	if (field[0]) {
		switch (field[0]) {
		case 'i':
		case 'o':
			xs->slant = 1;
			break;
		}
	}
	
	/* rate the weight value
	 */
	field[0] = (char) 0;
	xs->weight = 2;                     /* medium, in guesses below */
	readField(xs->xlfd, WEIGHT_NAME, field);
	if (field[0]) {
		if (!strncmp(field, "delicate", 8)) {
			xs->weight = 0;
		}
		else if (!strncmp(field, "light", 5)) {
			xs->weight = 0;
		}
		else if (!strncmp(field, "book", 4)) {
			xs->weight = 1;
		}
		else if (!strncmp(field, "medium", 6)) {
			xs->weight = 2;
		}
		else if (!strncmp(field, "demi", 4)) {
			xs->weight = 3;
		}
		else if (!strncmp(field, "bold", 4)) {
			xs->weight = 4;
		}
		else if (!strncmp(field, "black", 5)) {
			xs->weight = 5;
		}
	}
}

/* : splitFams - produce a finalized set of families
 * from the passed tentative family grouping.
 */
static void
splitFams(infamName, infam, outFams)
char *infamName;
xfamily *infam;
axhtab *outFams;
{
	int ix, jx, kx, itmp, sorted;
	int slants[4], weights[4], numvals;
	char outName[MAX_XLFD];
	xfamily outfam, *famp;
	xlfd_style xtmp;
	
	if (infam->numXlfds == 1) {
		formFamilyName(infam->xlfds[0].xlfd, TRUE, outName);
		axhGetItem(outName, outFams, (char *) NULL, &famp);  /* alloc */
		famp->numXlfds = 1;                                  /* init */
		famp->xlfds = (xlfd_style *) malloc(sizeof(xlfd_style));
		famp->xlfds[0] = infam->xlfds[0];
		return;
	}

	/* sort the members into increasing-strokeweight order
	 */
	itmp = infam->numXlfds - 1;
	sorted = FALSE;
	while (!sorted) {
		sorted = TRUE;  /* we hope */
		for (ix = 0; ix < itmp; ix++) {
			jx = ix + 1;
			if (infam->xlfds[ix].weight > infam->xlfds[jx].weight) {
				xtmp = infam->xlfds[jx];
				infam->xlfds[jx] = infam->xlfds[ix];
				infam->xlfds[ix] = xtmp;
				sorted = FALSE;
			}
		}
	}

	/* Try to form a 4-grouping selectable as {on,off} x {bold/ital}
	 * This is the preferred grouping for our applications.  Any members
	 * which don't fall into such a grouping become distinct families
	 */
	for (ix = 0; ix < infam->numXlfds; ) {
		
		if (infam->numXlfds - ix < 4) {
			goto reject;
		}
		
		/* look forward as far as possible to view the current
		 * element as in a tentative grouping with the following three;
		 * record the grouping's slants and weights
		 */
		for (jx = ix, kx = 0; kx < 4; jx++, kx++) {
			slants[kx] = infam->xlfds[jx].slant;
			weights[kx] = infam->xlfds[jx].weight;
		}
		
		/* reject unless for both weight and slant,
		 * there are exactly two distinct values
		 */
		numvals = 1;
		if (weights[1] != weights[0]) {
			numvals++;
		}
		if ((weights[2] != weights[0]) && (weights[2] != weights[1])) {
			numvals++;
		}
		if ((weights[3] != weights[0]) && (weights[3] != weights[1]) &&
			(weights[3] != weights[2])) {
			numvals++;
		}
		if (numvals != 2) {
			goto reject;
		}
		numvals = 1;
		if (slants[1] != slants[0]) {
			numvals++;
		}
		if ((slants[2] != slants[0]) && (slants[2] != slants[1])) {
			numvals++;
		}
		if ((slants[3] != slants[0]) && (slants[3] != slants[1]) &&
			(slants[3] != slants[2])) {
			numvals++;
		}
		if (numvals != 2) {
			goto reject;
		}
		
		/* at this point, starting from the current infam->xlfds element and
		 * looking 3 elements forward, weight vals should go: w1,w1,w2,w2.
		 * reject unless for each weight val, there are different slants.
		 */
		if (infam->xlfds[ix].slant == infam->xlfds[ix + 1].slant) {
			goto reject;
		}
		if (infam->xlfds[ix + 2].slant == infam->xlfds[ix + 3].slant) {
			goto reject;
		}
		/* accept infam->xlfds elements ix thru ix+3 as a single family;
		 * this family is fully selectable using on/off values of bold/ital.
		 */
		formFamilyName(infam->xlfds[0].xlfd, TRUE, outName);
		axhGetItem(outName, outFams, (char *) NULL, &famp);  /* alloc */
		famp->numXlfds = 4;                                  /* init */
		famp->xlfds = (xlfd_style *) malloc(4 * sizeof(xlfd_style));
		for (jx = 0; jx < 4; jx++) {
			famp->xlfds[jx] = infam->xlfds[ix + jx];
		}
		ix += 4;    /* loop var */
		continue;
			
reject:
		/* split infam->xlfds[ix] out as its own family.  Use the
		 * dash-separated style information in infam's name to generate
		 * a distinct family name
		 */
		formFamilyName(infam->xlfds[ix].xlfd, FALSE, outName);
		axhGetItem(outName, outFams, (char *) NULL, &famp);  /* alloc */
		famp->numXlfds = 1;                                  /* init */
		famp->xlfds = (xlfd_style *) malloc(sizeof(xlfd_style));
		famp->xlfds[0] = infam->xlfds[ix];
		ix++;
	}
}

static void
formFamilyName(inName, basic, outName)
char *inName;    /* xlfd from which to form name */
int basic;       /* bool: whether name can be formed from just FAMILY_NAME */
char *outName;   /* write onto caller's MAX_XLFD-sized memory */
{
	char field[MAX_XLFD];
	
	field[0] = 0;
	readField(inName, FAMILY_NAME, field);
	strcpy(outName, field);
	if (!basic) {
		/* now concat the weight, slant, and setwidth info
		 */
		field[0] = (char) 0;
		readField(inName, WEIGHT_NAME, field);
		if (field[0]) {
			strcat(outName, " ");
			strcat(outName, field);
		}

		field[0] = (char) 0;
		readField(inName, SLANT, field);
		if (field[0]) {
			switch (field[0]) {
				
			case 'i':
				strcat(outName, " Italic");
				break;
				
			case 'o':
				strcat(outName, " Oblique");
				break;
				
			}
		}
		
		field[0] = (char) 0;
		readField(inName, SETWIDTH_NAME, field);
		if (strcmp(field, "normal")) {
			strcat(outName, " ");
			strcat(outName, field);
		}
	}

	capWords(outName, strlen(outName));
}

static void
readEncodingAndMetricsNames(xlfd, filedata, encFiles, frec, f_xlfd)
char *xlfd;                  /* in */
axhtab *filedata;            /* in */
default_encodings *encFiles; /* in */
fontTabEnt *frec;            /* out: init encodingName, metricsNames members */
fontfile_xlfd **f_xlfd;      /* out: return ref corresponding to passed xlfd */
{
	int itmp;
	char *cp, *metricsPath;
	FILE *metFile;
	fontfile_xlfd *xf;

	/* lookup fontfile_xlfd instance associated with passed xlfd
	 */
	axhGetItem(xlfd, filedata, (char *) 0, (char **) &xf);
	*f_xlfd = xf;
	
	/* the native encoding file name is taken from the font filename suffix;
	 * typically a good guess, but no real guarantee..
	 */
	cp = xf->filename;
	while (*cp && *cp != '.') {
		cp++;
	}
	if (*cp && (strlen(cp) == 4)) {
		if (!strcmp(cp, ".ttf")) {
			frec->internalEncodingName = encFiles->truetype;
			frec->use.internalEncodingName = TRUE;
			if (xf && xf->dir) {
				/* for truetypes, we can read whether the font file
				 * is in the latin1 encoding; else claim "unspecifed"
				 */
				int encoding_int;
				char *ttname;

				itmp = strlen(xf->dir) + strlen(xf->filename) + 2;
				ttname = (char *) malloc (itmp);
				strcpy(ttname, xf->dir);
				strcat(ttname, "/");
				strcat(ttname, xf->filename);
				if (axttFontEncoding(ttname, &encoding_int) >= 0) {
					if (encoding_int == AXENC_UNSPECIFIED) {
						frec->internalEncodingName = UNSPECIFIED;
						
						/* supply an encodingFileName of UNSPECIFIED, also
						 */
						frec->encodingFileName =
							(char *) malloc (strlen(UNSPECIFIED) + 1);
						strcpy(frec->encodingFileName, UNSPECIFIED);
					}
				}
				free (ttname);
			}
		}
		if (!strcmp(cp, ".spd")) {
			frec->internalEncodingName = encFiles->speedo;
			frec->use.internalEncodingName = TRUE;
		}
		else if (!strcmp(cp, ".pfb") || !strcmp(cp, ".pfa")) {
			frec->internalEncodingName = encFiles->type1;
			frec->use.internalEncodingName = TRUE;
		}
	}
	
	/* the metrics filename for gallium font records is the xlfd,
	 * but scaled to AXGFS_TARGET_POINTSIZE
	 */
	itmp = strlen(xlfd) + 20; /* plus room for ptsize,xres,yres,null */
	frec->numTargetMetricsNames = 1;
	frec->targetMetricsNames = (char **) malloc(sizeof(char *));
	frec->targetMetricsNames[0] = malloc(itmp);
	axtFormScalableName(xlfd, AXGFS_TARGET_POINTSIZE,
						axRealDisplayResolution(), (axctm *) 0,
						frec->targetMetricsNames[0]);
	frec->use.numTargetMetricsNames = frec->use.targetMetricsNames = TRUE;

	frec->sizeEM = 1000;
	frec->use.sizeEM = TRUE;
	return;
}

/*=================*/

/* : newFontfileXLFD - used to create items in
 * axgfsReadFonts:ps.cg->screenNames
 */
static char *
newFontfileXLFD(xlfd, namedata)
char *xlfd;
char *namedata;
{
	char *cp;
	fontfile_xlfd *xf = (fontfile_xlfd *) namedata;
	fontfile_xlfd *newxf;
	
	newxf = (fontfile_xlfd *) malloc(sizeof(fontfile_xlfd));
	memset((void *) newxf, 0, sizeof(fontfile_xlfd));
	newxf->dir = (char *) malloc(strlen(xf->dir) + 1);
	strcpy(newxf->dir, xf->dir);
	newxf->filename = (char *) malloc(strlen(xf->filename) + 1);
	strcpy(newxf->filename, xf->filename);
	newxf->xlfd_template = (char *) malloc(strlen(xlfd) + 1);
	strcpy(newxf->xlfd_template, xlfd);
									 
	return ((char *) newxf);
}

/* : newFontfileXLFD - used to delete items in
 * axgfsReadFonts:ps.cg->screenNames
 */
static void
deleteFontfileXLFD(namedata, notUsed)
char *namedata;
char *notUsed;
{
	if (namedata) {
		fontfile_xlfd *xf = (fontfile_xlfd *) namedata;
		if (xf->xlfd_template) {
			free(xf->xlfd_template);
		}
		if (xf->dir) {
			free(xf->dir);
		}
		if (xf->filename) {
			free(xf->filename);
		}
		free(xf);
	}
}

/* : newFam - used to create items in axgfsReadFonts:fams
 */
static char *
newFam(fam, notUsed)
char *fam;
char *notUsed;
{
	xfamily *newfam = (xfamily *) malloc(sizeof(xfamily));
	memset((void *) newfam, 0, sizeof(xfamily));
	return ((char *) newfam);
}

/* : deleteFam - used to delete items in axgfsReadFonts:fams
 */
static void
deleteFam(fam, notUsed)
char *fam;
char *notUsed;
{
	xfamily *xf = (xfamily *) fam;
	if (xf) {
		if (xf->xlfds) {
			free(xf->xlfds);
		}
		free(xf);
	}
}

static char *
printError(part1, part2, part3, part4, part5)
char *part1;
char *part2;
char *part3;
char *part4;
char *part5;
{
	char *message, *mp;
	int len = 0;

	if (part1) {
		len += strlen(part1) + 1;
	}
	if (part2) {
		len += strlen(part2) + 1;
	}
	if (part3) {
		len += strlen(part3) + 1;
	}
	if (part4) {
		len += strlen(part4) + 1;
	}
	if (part5) {
		len += strlen(part5) + 1;
	}

	message = (char *) malloc(len);
	mp = message;
		
	if (part1) {
		strcpy(mp, part1);
		mp += strlen(part1);
	}
	if (part2) {
		strcpy(mp, part2);
		mp += strlen(part2);
	}
	if (part3) {
		strcpy(mp, part3);
		mp += strlen(part3);
	}
	if (part4) {
		strcpy(mp, part4);
		mp += strlen(part4);
	}
	if (part5) {
		strcpy(mp, part5);
	}
	
	return (message);
}

