/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: n_style.c
   Applix ASC version: 20
   Date: Fri Mar 19 07:42:54 1999
*/
#include <math.h>
#include <aos.h>
#include <gstyle.h>
#include <n_style.h>

/* time interval info (number of seconds in 24 hours) */
#define TIME_GRANULE (double)(1.0/(24*60*60*1000))
#define HH_FACTOR 3600000.0
#define MM_FACTOR 60000.0
#define SS_FACTOR 1000.0
#define MAXDATE 73414

/* maximum string lengths */
#define MAXDIGITS 60
#define DIGITS_AFTERDEC 9
#define MAXGENFORMATVAL ((double)(1.0E+15))

/* some simple tables */
static int PreviousDays[] = {
	/* Jan */ 0,
	/* Feb */ 31,
	/* Mar */ 31+28,
	/* Apr */ 31+28+31,
	/* May */ 31+28+31+30,
	/* Jun */ 31+28+31+30+31,
	/* Jul */ 31+28+31+30+31+30,
	/* Aug */ 31+28+31+30+31+30+31,
	/* Sep */ 31+28+31+30+31+30+31+31,
	/* Oct */ 31+28+31+30+31+30+31+31+30,
	/* Nov */ 31+28+31+30+31+30+31+31+30+31,
	/* Dec */ 31+28+31+30+31+30+31+31+30+31+30
};
static double PrecisionTable[] = {
	0.5,
	0.05,
	0.005,
	0.0005,
	0.00005,
	0.000005,
	0.0000005,
	0.00000005,
	0.000000005,
	0.0000000005,
	0.00000000005,
	0.000000000005,
	0.0000000000005,
	0.00000000000005,
	0.000000000000005,
	0.0000000000000005
};
static double PowersOfTenTable[] = {
	0.0,
	10.0,
	100.0,
	1000.0,
	10000.0,
	100000.0,
	1000000.0,
	10000000.0,
	100000000.0,
	1000000000.0
};


/******************************************************************************
 *
 * do_quick_float_format - number formatting utility
 */
do_quick_float_format(val, fmt, n_radix, str, max_chars)
double val;
number_format *fmt;
int n_radix;
char *str;
int max_chars;
{
	char *p, frac_str[STRPATH];
	double fractional_part;
	int integral_part, len;

	/* get the integral part of the number */
	p = str;
	if ( val >= 0.0 )
		integral_part = floor(val);
	else
	{
		integral_part = ceil(val);
		if ( integral_part == 0 )
			*p++ = '-';
	}
	sprintf(p, "%d", integral_part);

	/* get the fractional part */
	if ( n_radix > 0 )
	{
		/* add the decimal seperator */
		len = strlen(str);
		str[len] = '.';

		/* get the fractional part, formatting it as an integer */
		fractional_part = (val-integral_part) * PowersOfTenTable[n_radix];
		fractional_part	+= PrecisionTable[n_radix];
		
		integral_part = (int)floor(fabs(fractional_part));

		/* copy the fractional part into a temp string, padded with zeros */
		p = str + len+1;
		sprintf(frac_str, "%d", integral_part);
		len = strlen(frac_str);
		if ( len < n_radix )
		{
			memset(p, '0', n_radix - len);
			p += n_radix - len;
		}

		/* add the fraction to the string */
		strcpy(p, frac_str);
	}
}

/******************************************************************************
 *
 * do_basic_format - number formatting utility
 */
do_basic_format(val, fmt, str, max_chars, delete_trailers)
double val;
number_format *fmt;
char *str;
int max_chars, delete_trailers;
{
	int n_radix_places, len, n_digits, n_decimal_digits, got_dot, got_E;
	double absval;
	char *p, *pE;

	/* validate the numbers */
	if ( fmt->units )
		val = val / fmt->units;
	if ( dbl_equal(val, 0.0) )
		val = 0.0;
	n_radix_places = (fmt->n_radix_places >= 15) ? 9 : fmt->n_radix_places;

	/* round the value off if it's small enough */
	absval = fabs(val);
	if ( absval < (double)MAXINT )
	{
		if ( val < 0.0 )
			val -= PrecisionTable[n_radix_places];
		else
			val += PrecisionTable[n_radix_places];
		do_quick_float_format(val, fmt, n_radix_places, str, max_chars);
	}

	/* if it has more than 11 digits, force scientific notation */
	else if ( absval >= MAXGENFORMATVAL )
		sprintf(str,"%.*e", n_radix_places, val);

	/* just use regular sprintf formatting */
	else
		sprintf(str,"%.*lf", n_radix_places, val);

	/* now reformat the string we just created */
	n_digits = n_decimal_digits = 0;
	got_dot = got_E = FALSE;
	len = strlen(str);
	p = str;
	if ( *p == '-' )
		p++;
	if ( *p == '0' )
		p++;
	while ( *p )
	{
		/* an exponential marker? */
		if ( (*p == 'E') || (*p == 'e') )
			got_E = TRUE;

		/* a decimal point? */
		else if ( *p == '.' )
		{
			got_dot = TRUE;
			*p = fmt->d_sep;
		}

		/* assume a number */
		else
		{
			n_digits++;
			if ( got_dot && !got_E )
				n_decimal_digits++;

			if ( (n_digits>MAXDIGITS) || (n_decimal_digits>DIGITS_AFTERDEC) )
				*p = '0';
		}

		/* next character */
		p++;
	}

	/* remove trailing zeros if asked to */
	if ( delete_trailers && got_dot )
	{
		/* scientific notation? skip all digits until 'e' or 'E' */
		if ( got_E )
		{
			len--;
			while ( len >= 0 )
			{
				if ( str[len] == 'e' || str[len] == 'E' )
				{
					pE = str + len;
					break;
				}
				len--;
			}
		}

		/* find the last non-zero digit */
		while ( str[--len] == '0' );
		if ( str[len] != fmt->d_sep )
			len++;

		/* end the string */
		if ( got_E )
			strcpy(&str[len], pE);
		else
			str[len] = (char)0;
	}

	/* if there are no radix places, get rid of decimal point (if any) */
	if ( n_radix_places == 0 && got_dot && !got_E )
		str[len-1] = (char)0;

	/* get the final length */
	len = strlen(str);
	if ( len == 0 )
	{
		str[0] = '0';
		if ( n_radix_places <= 0 )
			len = 1;
		else
		{
			str[1] = fmt->d_sep;
			memset(&str[2], '0', n_radix_places);
			len = 2 + n_radix_places;
			str[len] = (char)0;
		}
	}
	return(len);
}

/******************************************************************************
 *
 * FmtNumCust - unstructured number format, uses fields in fmt struct only
 */
FmtNumCust(val, fmt, str, max_chars)
double val;
number_format *fmt;
char *str;
int max_chars;
{
	double min_val, power;
	int len, i, count, last_ix, dot_ix, com_ix;
	char *p, *p1, *str1;
	
	/* preprocess the number */
	if ( fmt->units > 0 )
	{	
		val = val / fmt->units;
		if ( fmt->n_radix_places > 0 )
		{
			if ( fmt->n_radix_places > 9 )
				fmt->n_radix_places = 9;
			power = fmt->n_radix_places * (-1.0);
			if ( (val >= 0.0 && val <= 1.0) && (power <= 0.0))
				min_val = 0.0;
			else
				min_val = pow(val, power);
			if ( fabs(val) < min_val )
				val = 0.0;
		}

		/* add any prefix */
		p = str;
		if ( fmt->prefix )
		{
			strcpy(str, fmt->prefix);
			len = strlen(str);
			p = str + len;
		}

		/* write the value into the string */
		sprintf(p, "%.*f", fmt->n_radix_places, val);
	}
	else
	{
		if ( fmt->n_radix_places > 9 )
			fmt->n_radix_places = 9;
		/* add any prefix */
		p = str;
		if ( fmt->prefix )
		{
			strcpy(str, fmt->prefix);
			len = strlen(str);
			p = str + len;
		}

		/* write the value into the string */
		sprintf(p, "%s", "Infinity");
	}

	/* add any t_sep: thousands separator */
	last_ix = strlen(str);
	if ( (str1 = strchr(str, 46)) != (char *)0 )
		dot_ix  = str1 - str;
	else
		dot_ix  = last_ix;
	com_ix  = dot_ix;
	count   = 0;
	while ( com_ix > 1 )
	{
		last_ix = strlen(str);
		if ( (str1 = strchr(str, 46)) != (char *)0 )
			dot_ix  = str1 - str;
		else
			dot_ix  = last_ix;
		com_ix = count>0 ? com_ix-3 : dot_ix-3;

		if ( com_ix > 1 )
		{
			p1 = &str[com_ix];
			memmove(&str[com_ix+1], &str[com_ix], 
					(size_t) (last_ix-com_ix + 1));
			count++;
			str[com_ix] = ',';
		}
	}
	/* substitute the correct seperators */
	len = strlen(str);
	for ( i=0; i<len; i++ )
	{
		if ( str[i] == '.' )
			str[i] = fmt->d_sep;
		else if ( str[i] == ',' )
			str[i] = fmt->t_sep;
	}

	/* remove trailing zeros, if needed */
	if ( fmt->trim_string )
	{
		len = strlen(str);
		while ( str[--len] == '0' );
		if ( str[len] == fmt->d_sep )
			len++;
		str[len] = (char)0;
	}
	
	/* add on a suffix */
	if ( fmt->suffix )
	{
		len = strlen(str);
		strcpy(&str[len], fmt->suffix);
	}
}

/******************************************************************************
 *
 * FmtNumNone
 */
FmtNumNone(val, fmt, str, max_chars)
double val;
number_format *fmt;
char *str;
int max_chars;
{
	int len;

	fmt->units = 1;
	fmt->n_radix_places = 9;
	len = do_basic_format(val, fmt, str, max_chars, TRUE);
}

/******************************************************************************
 *
 * FmtNumComm
 */
FmtNumComm(val, fmt, str, max_chars)
double val;
number_format *fmt;
char *str;
int max_chars;
{
	char tmpstr[STRPATH], *src, *p;
	int negative, next_comma;

	/* a paranoid sized default format */
	if ( max_chars > STRPATH )
		max_chars = STRPATH - STRPATH/3;
	else
		max_chars = max_chars - max_chars/3;
	do_basic_format(val, fmt, tmpstr, max_chars, fmt->trim_string);

	/* add an opening paren after the minus sign (if any) */
	src = tmpstr;
	negative = (*src == '-') ? TRUE : FALSE;
	if ( negative )
	{
		src++;
		*str++ = '(';
	}

	/* find the first decimal character */
	for ( p=src; (*p != fmt->d_sep) && (*p != (char)0); p++ );

	/* now index backwards, 3 at a time, looking for commas */
	next_comma = (p-src)%3;
	if ( next_comma == 0 )
		next_comma = 3;
	while ( src != p )
	{
		if ( next_comma-- == 0 )
		{
			*str++ = fmt->t_sep;
			next_comma = 2;
		}
		*str++ = *src++;
	}
	strcpy(str, src);

	/* now add the closing paren (if any) */
	if ( negative )
		strcat(str, ")");
}

/******************************************************************************
 *
 * FmtNumCurr
 */
FmtNumCurr(val, fmt, str, max_chars)
double val;
number_format *fmt;
char *str;
int max_chars;
{
	char buf[STRPATH], *p;
	int len, negative;

	if ( fmt->aux_id == FMT_MONEY1 ) /* $1,000.00 or ($1,000.00) */
	{
		/* prepend the prefix */
		if(fmt->prefix != NULL)
		{
			strcpy(str, fmt->prefix);
			len = strlen(fmt->prefix);
		}
		else
		{
			*str = '\0';
			len = 0;
		}

		/* format the number */
		FmtNumComm(val, fmt, &str[len], max_chars);

		/* place the paren before the prefix */
		if ( str[len] == '('  &&  fmt->prefix)
		{
			*str++ = '(';
			strncpy(str, fmt->prefix, len);
		}
	}
	else /* 1,000.00$ or (1,000.00$) */
	{
		/* format the number */
		FmtNumComm(val, fmt, str, max_chars);

		/* do we have a suffix */
		if ( fmt->suffix )
		{
			/* advance the string to the end */
			len = strlen(str);
			if ( *str != '(' )
				negative = FALSE;
			else
			{
				negative = TRUE;
				len--;
			}
			str += len;

			/* copy the suffix to the end */
			len = strlen(fmt->suffix);
			strncpy(str, fmt->suffix, len);
			str += len;

			/* add a closing paren if the number is negative */
			if ( negative )
				*str++ = ')';

			/* null terminate the string */
			*str = (char)0;
		}
	}
}

/******************************************************************************
 *
 * FmtNumFixd
 */
FmtNumFixd(val, fmt, str, max_chars)
double val;
number_format *fmt;
char *str;
int max_chars;
{
	int len, trim_string;
	trim_string = fmt->trim_string ? TRUE : FALSE;
	len = do_basic_format(val, fmt, str, max_chars, trim_string);
}

/******************************************************************************
 *
 * FmtNumGenl
 */
FmtNumGenl(val, fmt, str, max_chars)
double val;
number_format *fmt;
char *str;
int max_chars;
{
	int len;
	len = do_basic_format(val, fmt, str, max_chars, TRUE);
}

/******************************************************************************
 *
 * FmtNumScie
 */
FmtNumScie(val, fmt, str, max_chars)
double val;
number_format *fmt;
char *str;
int max_chars;
{
	int e_ix, d_ix, delta, back_ix;
	char *p, tmp[STRPATH];

	/* create the string, truncating the radix to 9 characters */
	p = str;
	if ( fmt->trim_string )
		sprintf(str, "%.*e", 9, val);

	/* create the string, rounding the mantissa */
	else
	{
		sprintf(str,"%.*e", fmt->n_radix_places, val);

		/* find the exponent symbol */
		for(;;)
		{
			if ( *p != 'e' && *p != 'E' )
				p++;
			else
				break;
		}

		/* copy the exponent into a temp buffer and terminate the mantissa */
		strcpy(tmp, p);
		*p = (char)0;

		/* round the mantissa */
		val = atof(str);
		do_basic_format(val, fmt, str, max_chars, fmt->trim_string);

		/* compse the mantissa and exponent */
		strcat(str, tmp);
	}

	/* read thru the string */
	for ( p=str; *p; p++ )
	{
		/* fix the decimal seperator when we find it */
		if ( *p == '.' )
		{
			d_ix = p - str;
			*p = fmt->d_sep;
		}

		/* note the position of the 'e' when found */
		if ( *p == 'e' )
		{
			e_ix = p - str;
			break;
		}
	}

	/* if instructed to trim zeros, do so */
	if ( fmt->trim_string )
	{
		/* find the last non-zero character after the allowed precision */
		back_ix = e_ix;
		while ( str[--back_ix] == '0' );
		if ( str[back_ix] != fmt->d_sep )
			back_ix++;

		/* reduce the string */
		strcpy(&str[back_ix], &str[e_ix]);
	}
}

/******************************************************************************
 *
 * FmtNumPcnt
 */
FmtNumPcnt(val, fmt, str, max_chars)
double val;
number_format *fmt;
char *str;
int max_chars;
{
	int len, units;
	val *= 100.0;
	/* save units : do_basic_format is using fmt->units info(WRONG!!)*/
	units = fmt->units;
	fmt->units = 1;
	len = do_basic_format(val, fmt, str, max_chars, fmt->trim_string);
	/* restore units */
	fmt->units = units;
	str[len] = '%';
	str[len+1] = (char)0;
}


/******************************************************************************
 *
 * FmtExtract - creates a time value in hours, min, sec and thousands of a sec
 *              helper routine for FmtNumDate
 */

FmtExtractTime(val, hh, mm, ss, tss)
double val;
int *hh;
int *mm;
int *ss;
int *tss;
{
	int granules;

	/* make sure the number is between 1.0 and -1.0 */
	val -= floor(val);
	if ( val < 0.0 )
		val += 1.0;

	/* get the hours */
	val = val / TIME_GRANULE;
	granules = (int)ceil(val);
	*hh = granules / HH_FACTOR;
	
	/* get the minutes */
	granules -= *hh * HH_FACTOR;
	*mm = granules / MM_FACTOR;

	/* get the seconds */
	granules -= *mm * MM_FACTOR;
	*ss = granules / SS_FACTOR;

	/* get the thousandth's of seconds */
	granules -= *ss * SS_FACTOR;
	*tss = granules;
	return(AOK);
}

/* This will do exactly the same as Formatdt, except the encoding/decoding
   that is done in Formatdt(), which was causing a 4second rounding -vamsi
 */  
typeErr
Formatdt_components(date, datefmt, retstrP)
DateStruct *date;
int		   datefmt;
mbchar	   *retstrP;		
{
	typeErr retval = AOK,
	lang = 1;			/*	default to English */
	int fullYr=FALSE;
	
	lang = id_lang();				/*	identify running language */
	Formatdt_lang_components(date->day, date->month, date->year, date->hour, date->minute, date->second, datefmt, retstrP, lang, fullYr);
	return(AOK);
}


/******************************************************************************
 *
 * FmtNumDate - creates a date from a number between 1 and MAXDATE (Dec 31,2100)
 */
FmtNumDate(dval, fmt, str, max_chars)
double dval;
number_format *fmt;
char *str;
int max_chars;
{
	register int leap_yr, n_yrs, n_days;
	unsigned long date_val;
	register long did;
	register int val;
	int hh, mm, ss, tss, lang;
	DateStruct date;

	/* set some initial value */
	date.year = 200;
	date.month = 11;
	date.day = 30;

	/* validate the number */
	val = floor(dval);
	if ( val > MAXDATE || val < 1 )
	{
		if ( fmt->display_format_errors )
			strcpy(str, "########");
		else
			do_basic_format(val, fmt, str, max_chars, TRUE);
	}

	/* must be a valid date */
	else
	{
		/* 0 ==> Jan 1, 1900, not 1 */
		date_val = val;
		date_val--;

		did = 366;	/* for the year 1900 */
		n_days = 366;	/* for 1900 */
		leap_yr = TRUE;

		/* count the days and years prior to the current date */
		for ( n_yrs=1; did<=date_val; n_yrs++ )
		{
			leap_yr = (0 == n_yrs % 4)- (0 == n_yrs % 100) + (100 == n_yrs);
			n_days = 365 + leap_yr;
			did += n_days;
		}
		did -= n_days;

		/* set the current year */
		date.year = n_yrs-1 + 1900 ;

		/* set the current month of the year */
		date_val -= did;
		/*for ( date.month=11; date.month>0; date.month-- ) */
		for ( date.month=11; ; date.month-- )
		{
			n_days = PreviousDays[date.month] + (leap_yr && (date.month > 1));
			if ( n_days <= date_val )
				break;
		}
		date.month++;

		/* set the current day of the week */
#if 0
		date.day = (n_days<=date_val) ? date_val-n_days+1 : date_val+1;
#else
		date.day = date_val - n_days + 1;
#endif

		/* clear the time */		
		date.hour = date.minute = date.second = 0;
		/* Format the time, if any??? */
		FmtExtractTime(dval, &hh, &mm, &ss, &tss);
		date.hour   = hh;
		date.minute = mm;		
		date.second = ss;
#if 1
		Formatdt_components(&date, fmt->aux_id, str);
#else
		date_val = Builddt(&date);
		Formatdt(date_val, fmt->aux_id, str);
#endif		
	}
}


/******************************************************************************
 *
 * FmtNumBool
 */
FmtNumBool(val, fmt, str, max_chars)
double val;
number_format *fmt;
char *str;
int max_chars;
{
	if ( dbl_equal( val, 0.0 ) )
		sprintf(str, "%s", fmt->false);
	else
		sprintf(str, "%s", fmt->true);
}

/******************************************************************************
 *
 * FmtNumGrph
 */
FmtNumGrph(val, fmt, str, max_chars)
double val;
number_format *fmt;
char *str;
int max_chars;
{
	/* make sure the number is between 0.0 and 200.0 */
	if ( val > 200.0 )
		val = 200.0;
	else if ( val < 0.0 )
		val = 0.0;

	/* set the graphics characters into the string */
	memset(str, '*', (int)val);
	str[(int)val] = '\0';
}

/******************************************************************************
 *
 * FmtNumTime - given a val between 1.0 and -1.0, generate time string
 */
FmtNumTime(val, fmt, str, max_chars)
double val;
number_format *fmt;
char *str;
int max_chars;
{
	int granules, hh, mm, ss, tss;

	/* make sure the number is between 1.0 and -1.0 */
	val -= floor(val);
	if ( val < 0.0 )
		val += 1.0;

	/* get the hours */
	val = val / TIME_GRANULE;
	granules = (int)ceil(val);
	hh = granules / HH_FACTOR;
	
	/* get the minutes */
	granules -= hh * HH_FACTOR;
	mm = granules / MM_FACTOR;

	/* get the seconds */
	granules -= mm * MM_FACTOR;
	ss = granules / SS_FACTOR;

	/* get the thousandth's of seconds */
	granules -= ss * SS_FACTOR;
	tss = granules;

	/* format the string */
	switch ( fmt->aux_id )
	{
		default:
		case FMT_TIME1: /* hh:mm:ss AM/PM (12 hour clock) */
			if ( hh < 12 )
			{
				if ( hh <= 0 )
					hh = 12;
				sprintf(str, "%02d:%02d:%02d %s", hh, mm, ss, fmt->am);
			}
			else
			{
				if ( hh > 12 )
					hh -= 12;
				sprintf(str, "%02d:%02d:%02d %s", hh, mm, ss, fmt->pm);
			}
		break;

		case FMT_TIME2: /* hh:mm AM/PM (12 hour clock) */
			if ( hh < 12 )
			{
				if ( hh <= 0 )
					hh = 12;
				sprintf(str, "%02d:%02d %s", hh, mm, fmt->am);
			}
			else
			{
				if ( hh > 12 )
					hh -= 12;
				sprintf(str, "%02d:%02d %s", hh, mm, fmt->pm);
			}
		break;

		case FMT_TIME3: /* hh:mm:ss (24 hour clock) */
			sprintf(str, "%02d:%02d:%02d", hh, mm, ss);
		break;

		case FMT_TIME4: /* hh:mm (24 hour clock) */
			sprintf(str, "%02d:%02d", hh, mm);
		break;

		case FMT_TIME5: /* hh:mm:ss.tss AM/PM (12 hour clock) */
			if ( hh < 12 )
			{
				if ( hh <= 0 )
					hh = 12;
				sprintf(str, "%02d:%02d:%02d%c%03d %s",
				             hh, mm, ss, fmt->d_sep, tss,
				             fmt->am);
			}
			else
			{
				if ( hh > 12 )
					hh -= 12;
				sprintf(str, "%02d:%02d:%02d%c%03d %s",
				             hh, mm, ss, fmt->d_sep, tss,
				             fmt->pm);
			}
		break;

		case FMT_TIME6: /* hh:mm:ss.tss (24 hour clock) */
			sprintf(str, "%02d:%02d:%02d%c%03d", hh, mm, ss, fmt->d_sep, tss);
		break;
	}
}

/******************************************************************************
 *
 * FormatNumber - Top level number formatter
 */
FormatNumber(val, in_fmt, str, max_chars)
double val;
number_format *in_fmt;
char *str;
int max_chars;
{
	number_format fmt;
	fmt = *in_fmt;
	if ( fmt.n_radix_places > 9 )
		fmt.n_radix_places = 15;
	switch ( fmt.id )
	{
		default:
		case FMT_CUSTOM:     FmtNumCust(val, &fmt, str, max_chars); break;

		case FMT_UNSTYLED:   FmtNumNone(val, &fmt, str, max_chars); break;
		case FMT_CURRENCY:   FmtNumCurr(val, &fmt, str, max_chars); break;
		case FMT_COMMA:      FmtNumComm(val, &fmt, str, max_chars); break;
		case FMT_FIXED:      FmtNumFixd(val, &fmt, str, max_chars); break;
		case FMT_GENERAL:    FmtNumGenl(val, &fmt, str, max_chars); break;
		case FMT_SCIENTIFIC: FmtNumScie(val, &fmt, str, max_chars); break;
		case FMT_PERCENTAGE: FmtNumPcnt(val, &fmt, str, max_chars); break;
		case FMT_DATE:       FmtNumDate(val, &fmt, str, max_chars); break;
		case FMT_BOOLEAN:    FmtNumBool(val, &fmt, str, max_chars); break;
		case FMT_GRAPH:      FmtNumGrph(val, &fmt, str, max_chars); break;
		case FMT_TIME:       FmtNumTime(val, &fmt, str, max_chars); break;
	}
}
