/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axselect.c
   Applix ASC version: 42
   Date: Mon Nov 16 13:53:27 1998
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <axtime.h>
#include <signal.h>
#include <axsocket.h>
#include <axthread.h>
#if !AXMS
#include <sys/param.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <arpa/inet.h>
#endif
#if RIOS
#include <sys/select.h>
#endif
#include <errno.h>
#include <axnet.h>
#define THIM_STATIC 1    /** do not redefine such things as "exit()"  */
#include <aos.h>
#include <elfapi.h>
#if !ELF_COMM_STAND_ALONE
#include <thim_gen.h>
#endif

#define boolean int

/*
	This file provides a robust callback based select() layer for
	standalone (non-axmain) program IO.

	This file is divided into "Convenience" and "Core" functions.
	The convenience functions exist to avoid having the consumer directly
	modify structure contents.

	The axio_channel_t creation/destruction is thread locked with
	a mutex.  The members of created axio_channel_t's are not thread
	locked because we assume only one thread modified them.
	The "for ( chanP = AllChannels; ..." loops are threadlocked
	only if missing a channel would be harmful.
*/
#include <axselect.p>

extern int axErrorCode;
extern char *axErrorString;
extern char *axErrorObject;
extern jmp_buf axErrorHandler;
extern int AxnetNoResponse;



/********************** CONVENIENCE FUNCTIONS *****************************/

/*************************************************************************
function: AxIORegisterService - Add server socket to select list
When a conneciton is made to the passed channel, the onAccept callback
is called, passing the axio_channel_t.
*/
axio_channel_t *AxIORegisterService (
	int channel,		/* from socket(), already bind()'d & listen()'d */
	void (*onAccept)(axio_channel_t *,struct sockaddr_in*))/* routine to call on client connects */
{
	axio_channel_t *chanP;
	chanP = AxIOAddConnectionInfo();
	chanP->sock_fd = channel;
	chanP->acceptFunc = onAccept;
	chanP->accepting = TRUE;
	return(chanP);
}

/*************************************************************************
function: AxIOAcceptClient - Allow connection from incoming client
This should only be called by the "onAccept" callback.
*/
axio_channel_t *AxIOAcceptClient (
	axio_channel_t *chanP, /* listening socket, as per AxIORegisterService */
	int (*readFunc)(axio_channel_t *,char *,int),	/* routine to call when data read from client */
	void (*writFunc)(axio_channel_t *,char *, int),	/* optional call after writing data to func */
	void (*closFunc)(axio_channel_t *))	/* optional call when shutting down connection */
{
	axio_channel_t *clientP;

	if ( !chanP->accepting || (chanP->chan_fd == -1) )
		return(NULL);
	clientP = AxIOAddConnectionInfo();
	if ( !clientP )
		return(NULL);
	clientP->closeNtfy = closFunc;
	clientP->chan_fd = chanP->chan_fd;
	clientP->readFunc = readFunc;
	clientP->reading = TRUE;
	clientP->writeNtfy = writFunc;
	clientP->writing = TRUE;
	return(clientP);
}

/*************************************************************************
function: AxIOAddDuplexChannel - Add a channel to select list for read & write
*/
axio_channel_t *AxIOAddDuplexChannel (
	int channel,
	int (*readFunc)(axio_channel_t *,char *,int),	/* routine to call when data read from client */
	void (*writFunc)(axio_channel_t *,char *, int),	/* optional call after writing data to func */
	void (*closFunc)(axio_channel_t *))	/* optional call when shutting down connection */
{
	axio_channel_t *chanP;

	if ( channel < 0 )	/* would trigger crash on select(...-fd...) */
		return(NULL);
	AxIOMakeChannelNonBlocking(channel);
	chanP = AxIOAddConnectionInfo();
	chanP->closeNtfy = closFunc;
	chanP->chan_fd = channel;
	chanP->readFunc = readFunc;
	chanP->reading = TRUE;
	chanP->writeNtfy = writFunc;
	chanP->writing = TRUE;
	return(chanP);
}

/*************************************************************************
function: AxIOSetClientInfo - Set the CallerInfo member
*/
void AxIOSetClientInfo (axio_channel_t *chanP,
	void *anyData)
{
	chanP->calledInfo = anyData;
}

/*************************************************************************
function: AxIOGetClientInfo - Get the CallerInfo member
*/
void *AxIOGetClientInfo (axio_channel_t *chanP)
{
	return(chanP->calledInfo);
}

/*************************************************************************
function: AxIOSetAcceptFunc - Note new acceptFunc
*/
void AxIOSetAcceptFunc (axio_channel_t *chanP,
	void (*acceptFunc)(axio_channel_t *,struct sockaddr_in*)) {
	chanP->acceptFunc = acceptFunc;
}

/*************************************************************************
function: AxIOSetWriteFunc - Note write completion
*/
void AxIOSetWriteFunc (axio_channel_t *chanP,
    void (*writeFunc)(axio_channel_t *,char *, int) )
{
    chanP->writeNtfy = writeFunc;
}



/*************************************************************************
function: AxIOIsClosing - Return whether channel tagged for shutdown
This becomes true whenever a read or write detects broken pipe.
*/
int AxIOIsClosing (axio_channel_t *chanP)
{
	return(chanP->channelClosed);
}

/*************************************************************************
function: AxIOCloseChannel - Tag channel for closing
*/
void AxIOCloseChannel (axio_channel_t *chanP)
{
	if ( !chanP )
		return;
	chanP->channelClosed = TRUE;
	chanP->reading = FALSE;
	chanP->writing = FALSE;
	chanP->accepting = FALSE;
}

/*************************************************************************
function: AxIOSoftCloseChannel - Perform shutdown-less close
Close and tag channel now.  This prevents channel from being shutdown
by the ShutDownConnection() routine.  Needed for servers that start
other processes (like aagtwy) that need to do a 'softer' channel close.
*/
void AxIOSoftCloseChannel (axio_channel_t *chanP)
{
	if ( !chanP )
		return;
	if ( chanP->chan_fd != -1 ) {
		CLOSE_CHANNEL(chanP->chan_fd);
		chanP->chan_fd = -1;
	}
	chanP->channelClosed = TRUE;
	chanP->reading = FALSE;
	chanP->writing = FALSE;
	chanP->accepting = FALSE;
}

/*************************************************************************
function: AxIORejectClient - Don't accept client connection
Sends optional rejection message and closes connection.
This is called instead of 
*/
void AxIORejectClient (axio_channel_t *chanP,
		char *msgBuf,		/* optional rejection message */
		int msgLen)			/* length of rejection message */
{
	if ( !chanP->accepting )	/* really an error condition... */
		return;
	AxIOAddToWriteBuffer(chanP,msgBuf,msgLen,TRUE);
}

/********************** ELF SVR/CLI FUNCTIONS ******************************/
#if ELF_COMM_STAND_ALONE
extern int AxPreventLowCodesFromClient;
extern elfData AxDfltDispatch();

/*************************************************************************
function: AxIOElfSvrAccept - Accept connection, set read/write callbacks
*/
void AxIOElfSvrAccept (axio_channel_t *chanP, struct sockaddr_in *inP) {
	axio_channel_t *clientP;
	clientP = AxIOAcceptClient(chanP,AxIOElfReadFromClient,NULL,NULL);
}

/*************************************************************************
function: AxIOIsElfMsgComplete - Returns msgSize of FIRST msg if there is at least one
*/
int AxIOIsElfMsgComplete (char *buf, int cnt) {
	int32 inSize;
	int bufdex = 0;

	if ( cnt < sizeof(int32) )
		return(0);
	memcpy(&inSize,buf+bufdex,sizeof(int32));
	inSize = ntohl(inSize) + sizeof(int32);
	return( inSize > cnt ? 0 : inSize );
}

/*************************************************************************
function: AxIOIsDataComplete - Returns msgSize of ALL msgs if all are completed
*/
int AxIOIsDataComplete (char *buf, int cnt) {
	int32 inSize;
	int bufdex = 0;
	while (bufdex < cnt) {
		if (cnt < bufdex+sizeof(int32))
			break;
		memcpy(&inSize,buf+bufdex,sizeof(int32));
		inSize = ntohl(inSize);
		if (inSize > cnt-bufdex-sizeof(int32))
			break;
		bufdex += inSize+sizeof(int32);
	}
	return (bufdex == cnt) ? cnt : 0;
}

/*************************************************************************
function: AxIOElfReadFromClient - Process incoming ELf message - AxDispatch
*/
int AxIOElfReadFromClient (axio_channel_t *chanP, char *buf, int cnt) {
    elfData aP;     /* passed to function */
	int32 inSize;

/* get size, if whole msg not here here, ignore it */

	if ( 0 == (inSize = AxIOIsElfMsgComplete(buf,cnt)) )
		return(0);

/* convert to ELF data */

   	aP = AxRPCRead(buf+sizeof(int32));

/* call dispatcher */

	AxIODispatchElfMsg(chanP,aP);
    AxFreeData(aP);
	return(inSize);
}


void AxIOSendElfMsg (axio_channel_t *chanP, elfData rP)
{
    int32 size;     /* size of socket message */
    char *outBuf;

    size = AxMemWritePacked(rP,0,&outBuf,chanP->axio_flags.io_packed);
    if ( !AxnetNoResponse )
		AxIOAddToWriteBuffer(chanP,outBuf,size,FALSE);
    AxnetNoResponse = FALSE;

/* Release arrays and temporarily allocated data */

    AxFreeData(rP);
    AxFreeTempData();
    AxFree(outBuf);
    return;
}


/*************************************************************************
function: AxIODispatchElfMsg - Call customer supplied dispatcher
*/
void AxIODispatchElfMsg (axio_channel_t *chanP, elfData aP) 
{
    elfData fP,     /* returned from function */
            rP,     /* sent back to Aster*x */
            errP;   /* on error, sent back to Aster*x */
    char *outBuf;   /* holds the message on the way out */
    int32 size;     /* size of socket message */
    int cmdCode;
    elfData subP;
    int senders_task;

/* call customer supplied dispatcher */

    if ( setjmp(axErrorHandler) ) {
        rP = AxMakeArray(3);
        rP = AxAddIntToArray(rP,0,1);
        rP = AxAddIntToArray(rP,2,senders_task);

        errP = AxMakeArray(3);
        errP = AxAddIntToArray(errP,0,axErrorCode);
        errP = AxAddStrToArray(errP,1,axErrorString);
        errP = AxAddStrToArray(errP,2,axErrorObject);
        rP = AxAddArrayToArray(rP,1,errP);
    }
    else {
        cmdCode = AxIntFromArray(aP,0);
        subP = AxArrayFromArray(aP,1);
        senders_task = AxIntFromArray(aP,2);

        if ( (cmdCode < 1) && AxPreventLowCodesFromClient )
            AxError(1,NXLT("Client sending invalid cmdCode to AxDispatch"),NULL);
        fP = AxDispatchMsg(chanP,cmdCode,subP);
        if ( !fP && (cmdCode < 0) )
            fP = AxDfltDispatchMsg(chanP,cmdCode,subP);

        /* respond to client */
        rP = AxMakeArray(3);
        rP = AxAddIntToArray(rP,0,0);
        rP = AxAddArrayToArray(rP,1,fP);
        rP = AxAddIntToArray(rP,2,senders_task);
    }
    if ( !AxnetNoResponse ) {
	    size = AxMemWritePacked(rP,0,&outBuf,chanP->axio_flags.io_packed);
		AxIOAddToWriteBuffer(chanP,outBuf,size,FALSE);
	    AxFree(outBuf);
	}
    AxnetNoResponse = FALSE;

/* Release arrays and temporarily allocated data */

    AxFreeData(rP);
    AxFreeTempData();
}
#endif /* ELF_COMM_STAND_ALONE */

/********************** CORE FUNCTIONS ************************************/

axio_channel_t *AllChannels;	/* mutex controlled linked list */
MUTEX_LOCK ChanListMutex;

static fd_set SavedReadFds, SavedWriteFds;
static int SavedFdsValid;
static int WrappingUp;

/*************************************************************************
function: AxIOSelect - Base select routine, all incoming done with callbacks
Returns only when timeout elapsed.
*/
int AxIOSelect (timeP)
struct timeval *timeP;
{
	int events;

	while ( TRUE ) {
		events = AxIOSelectOne(timeP);
		if ( events == 0 )
			return(0);
	}
}

/*************************************************************************
function: AxIOSelectOne - Perform a select, handle at most 1 callback
*/
int AxIOSelectOne (timeP)
struct timeval *timeP;
{
	int didCall;
	axio_channel_t *chanP, *next;
	int events;
	fd_set read_fds, write_fds;

	if ( SavedFdsValid ) {
		read_fds = SavedReadFds;
		write_fds = SavedWriteFds;
	}
	else {
		int fd_sets = 0;
		FD_ZERO(&read_fds);
		FD_ZERO(&write_fds);
		for ( chanP = AllChannels; chanP; chanP = chanP->next ) {
			if ( chanP->accepting && !WrappingUp && chanP->sock_fd > 0)
				FD_SET(chanP->sock_fd, &read_fds);
			if ( chanP->reading && chanP->chan_fd > 0)
				FD_SET(chanP->chan_fd, &read_fds);
			if ( chanP->writing && chanP->writeCnt && chanP->chan_fd > 0) {
				FD_SET(chanP->chan_fd, &write_fds);
				fd_sets++;
			}
		}
/*		SavedFdsValid = TRUE; */
		if ( WrappingUp && !fd_sets )
			return(0);
	}
	events = select(FD_SETSIZE, &read_fds, &write_fds, NULL, timeP);
	if ( !events )
		return(0);
	if ( events < 0 ) {
#if AXMS
		errno = WSAGetLastError();
		if ( errno == WSAEBADF || errno == WSAENOTSOCK )
#else
		if ( errno == EBADF )
#endif
			RemoveBadChannel();
		return(-1);  /* presumably ignore EINTR, restart after EBADF */
	}

	for ( chanP = AllChannels; chanP; chanP = next ) {
		didCall = FALSE;
		if ( chanP->accepting && FD_ISSET(chanP->sock_fd, &read_fds) ) {
			didCall = TRUE;
			AcceptConnection(chanP);
		}
		if ( chanP->reading && FD_ISSET(chanP->chan_fd, &read_fds) ) {
			didCall = TRUE;
			ReadSomeFromChannel(chanP);
		}
		if ( chanP->writing && chanP->writeCnt &&
									FD_ISSET(chanP->chan_fd, &write_fds) ) {
			didCall = TRUE;
			WriteSomeToChannel(chanP);
		}
		next = chanP->next;
		if ( chanP->channelClosed ) {
			didCall = TRUE;
			ShutDownConnection(chanP);
		}
		if ( didCall )
			break;
	}
	return(events);
}

/*************************************************************************
function: AxIOWriteSome - Write to ready write channels
*/
void AxIOWriteSome ()
{
	axio_channel_t *chanP, *next;
	int events;
	fd_set read_fds, write_fds;
	static struct timeval timezero;
	struct timeval *timeP = &timezero;


	while ( TRUE ) {
		FD_ZERO(&read_fds);
		FD_ZERO(&write_fds);
		for ( chanP = AllChannels; chanP; chanP = chanP->next ) {
			if ( chanP->writing && chanP->writeCnt && chanP->chan_fd > 0)
				FD_SET(chanP->chan_fd, &write_fds);
		}
		events = select(FD_SETSIZE, &read_fds, &write_fds, NULL, timeP);
		if ( !events )
			return;
		if ( events < 0 ) {
#if AXMS
			errno = WSAGetLastError();
			if ( errno == WSAEBADF || errno == WSAENOTSOCK )
#else
			if ( errno == EBADF )
#endif
				RemoveBadChannel();
			return;  /* presumably ignore EINTR, restart after EBADF */
		}
	
		for ( chanP = AllChannels; chanP; chanP = next ) {
			if ( chanP->writing && chanP->writeCnt && FD_ISSET(chanP->chan_fd, &write_fds) )
				WriteSomeToChannel(chanP);
			next = chanP->next;
			if ( chanP->channelClosed )
				ShutDownConnection(chanP);
		}
	}
}

/*************************************************************************
function: AxIOWrapUp - Flush write buffers and return from AxIOSelect()
Disables accepts, reads.  Sends Q'd writes and allows AxIOSelect() to return(0)
*/
void AxIOWrapUp ()
{
	WrappingUp = TRUE;
}

/*************************************************************************
function: RemoveBadChannel - Find & shut down the channel that went bad
This should never be called...
Returns TRUE if found a bad channel.
Quits after tossing first bad one.  Relies on caller to call back in if there
are more bad ones.
*/
static int RemoveBadChannel ()
{
    fd_set read_fds, write_fds;
    struct timeval timeout;
    int events;
	axio_channel_t *chanP;

    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

	for ( chanP = AllChannels; chanP; chanP = chanP->next ) {
		FD_ZERO(&read_fds);
		FD_ZERO(&write_fds);
		if ( chanP->accepting && chanP->sock_fd > 0)
			FD_SET(chanP->sock_fd, &read_fds);
		if ( chanP->reading && chanP->chan_fd > 0)
			FD_SET(chanP->chan_fd, &read_fds);
		if ( chanP->writing && chanP->chan_fd > 0)
			FD_SET(chanP->chan_fd, &write_fds);
		events = select(FD_SETSIZE, &read_fds, &write_fds, NULL, &timeout);
		if ( events < 0 ) {
#if AXMS
        	errno = WSAGetLastError();
        	if ( errno == WSAEBADF || errno == WSAENOTSOCK ) {
#else
			if ( errno == EBADF ) {
#endif
				ShutDownConnection(chanP);
				return(TRUE);
			}
		}
	}
    return(FALSE);
}

/*************************************************************************
function: AcceptConnection - Accept connection and callback
Returns TRUE if removed connection.
If accepting socket wakes up, accept connection and load client info
info the read buffer.  This is where the read buffer is created.
The caller may (1) reset the 'accepting' flag and re-use this channel
for read/writes or (2) create a new channel, copying the chan_fd.
If the called routine doesn't like the client, the called routine
simply sets the chanP->channelClosed member.
If the called routine wants to send a reject, it needs to load the
write buffer and set the closeOnWrite member.

Note that while accept's second arg is formally an arbitrary sockaddr,
we presume TCP connections and thus only expect sockaddr_in's.
*/
static
void AcceptConnection (axio_channel_t *chanP)
{
	int clilen;
	struct sockaddr_in in_cli_addr;

	clilen = sizeof(in_cli_addr);
	chanP->chan_fd = accept(chanP->sock_fd,(struct sockaddr*)&in_cli_addr,&clilen);
	if ( chanP->chan_fd != -1 ) {
		AxIOMakeChannelNonBlocking(chanP->chan_fd);
		(*chanP->acceptFunc)(chanP,&in_cli_addr);
	}
}

/*************************************************************************
function: AxIOAddConnectionInfo - Allocate & thread a axio_channel_t
This is thread locked by ChanListMutex to prevent two threads from
creating/deleting at the same time.
*/
axio_channel_t *AxIOAddConnectionInfo ()
{
	axio_channel_t *chanP;
	static boolean ChanListMutexInited = FALSE;

	chanP = (axio_channel_t *)calloc(sizeof(axio_channel_t),1);
	if ( !chanP )
		return(NULL);
	LOCK_MUTEX(ChanListMutex);
	chanP->next = AllChannels;
	AllChannels = chanP;
	UNLOCK_MUTEX(ChanListMutex);
	return(chanP);
}

static
int vfyLiveChannel(axio_channel_t *closeP) {
	axio_channel_t *chanP;
	for ( chanP = AllChannels; chanP; chanP = chanP->next )
		if ( chanP == closeP )
			return(TRUE);
	return(FALSE);
}

/*************************************************************************
function: AxIOHackShutdownForEntsvr - TEMP!
*/
void AxIOHackShutdownForEntsvr (axio_channel_t *closeP)
{
	if ( closeP->chan_fd > 0 )
		ShutDownConnection(closeP);
}

/*************************************************************************
function: ShutDownConnection - Close out an accept()'d connection
This is thread locked by ChanListMutex to prevent two threads from
creating/deleting at the same time.
*/
static
void ShutDownConnection (axio_channel_t *closeP)
{
	axio_channel_t *chanP, *prev;

	/* make sure not already shut - yes, this is an issue for proxy-ers */
	if ( !vfyLiveChannel(closeP) )
		return;

	if ( closeP->chan_fd != -1 ) {
		shutdown(closeP->chan_fd,2);
		CLOSE_CHANNEL(closeP->chan_fd);
	}
	if ( closeP->closeNtfy )
		(*closeP->closeNtfy)(closeP);
	LOCK_MUTEX(ChanListMutex);
	if ( closeP == AllChannels )
		AllChannels = AllChannels->next;
	else {
		for ( chanP = AllChannels->next, prev = AllChannels;
			chanP && chanP != closeP;
			prev = chanP, chanP = chanP->next );
		if ( chanP != NULL )
			prev->next = closeP->next;
	}
	closeP->freed = TRUE;
	if ( closeP->readBuf )
		free(closeP->readBuf);
	if ( closeP->writeBuf )
		free(closeP->writeBuf);
	free(closeP);
	UNLOCK_MUTEX(ChanListMutex);
}

/*************************************************************************
function: AxIOProcessDeferred - Allows client to now process incoming message that was deferred
This is used if a prior readFunc returned 0 even though a complete msg was
buffered.  Needed because the AxSelect will *not* wake the read up again,
the message has already been buffered.
*/
void AxIOProcessDeferred (axio_channel_t *chanP) {
	int avail, did;

	avail = chanP->readPos - chanP->readOffset;
	did = (*chanP->readFunc)(chanP,chanP->readBuf+chanP->readOffset,avail);
	if ( (did < 0) || (did > avail) )
		did = avail;		/* protection from client returning junk */
	if ( did == avail ) {
		chanP->readPos = 0;
		chanP->readOffset = 0;
	}
	else {
		chanP->readOffset += did;
	}
}

/*************************************************************************
function: ReadSomeFromChannel - Read into realloced buffer, callback
If the read channel wakes up, realloc buffer to hold material
and attempt to read at least BASIC_BUF worth of bytes.
Shuts channel down on disconnect.
*/
static
void ReadSomeFromChannel (axio_channel_t *chanP)
{
	boolean haveDrained = FALSE;
	int did, avail, newSize;
	char *newBuf;
	int didThisCall = 0;	/* total read this call */
#define BASIC_BUF 8192	/* reasonable value, non-critical number */

	while ( !chanP->channelClosed && !haveDrained ) {
		avail = chanP->readBufSize - chanP->readPos;
		if ( avail < BASIC_BUF ) {
			newSize = 2 * (chanP->readBufSize + BASIC_BUF);
			if ( chanP->readBuf )
				newBuf = (char*)realloc(chanP->readBuf,newSize);
			else
				newBuf = (char*)malloc(newSize);
			if ( newBuf != NULL ) {	/* gracefully deal with no-mem */
				chanP->readBufSize = newSize;
				chanP->readBuf = newBuf;
			}
			avail = chanP->readBufSize - chanP->readPos;
			if ( avail == 0 )	/* still nothing, hope for better future! */
				return;
		}
#if AXMS
		did = recv(chanP->chan_fd, chanP->readBuf+chanP->readPos, avail, 0 );
		if ( did == SOCKET_ERROR ) {
			errno = WSAGetLastError();
			if ( (errno == WSAENETDOWN) || (errno == WSAESHUTDOWN) || 
				(errno == WSAEBADF) || (errno == WSAENOTSOCK) ||
				(errno == WSAECONNABORTED) || (errno == WSAECONNRESET) )
			   /* The calling code treats a 0 return from here to mean	 */
			   /* 'client went away'.  -1 is returned for other error conditions.*/
				chanP->channelClosed = TRUE;
			else if ( errno == WSAEWOULDBLOCK )
				haveDrained = TRUE;
		}
#else
		did = read(chanP->chan_fd, chanP->readBuf+chanP->readPos, avail );
		if ( did < 0 ) {
			/* Note: On SIEMENS/SINIX, the above read returns -1 with
			   errno is set to one of below when the client end does a
			   shut-down.  All other platforms return 0.
			   Therefore, sense the condition of the other side (client)
			   doing a shut-down, and return a 0.
			   The calling code treats a 0 return from here to mean
			   'client went away'.  -1 is returned for all other error
			   conditions. */
			if (errno == ECONNRESET || errno == ECONNABORTED || errno == EBADF)
				chanP->channelClosed = TRUE;
			else if (errno == EWOULDBLOCK)
				haveDrained = TRUE;
		}
#endif
		else if ( did == 0 ) {
			if ( didThisCall == 0 )	/* don't tag channel closed if data avail */
				chanP->channelClosed = TRUE;
			haveDrained = TRUE;
		}
		else {
			didThisCall += did;
			chanP->readPos += did;
			if ( did != avail )
				haveDrained = TRUE;		/* reasonable assumption */
		}
	}
	if ( chanP->channelClosed || (chanP->readPos >= chanP->readNeed) ) {
rdsome:	avail = chanP->readPos - chanP->readOffset;
		did = (*chanP->readFunc)(chanP,chanP->readBuf+chanP->readOffset,avail);
		if ( (did < 0) || (did > avail) )
			did = avail;		/* protection from client returning junk */
		if ( did == avail ) {
			chanP->readPos = 0;
			chanP->readOffset = 0;
		}
		else if ( did != 0 ) {	/* processed some of buffer, loop again */
			chanP->readOffset += did;
			goto rdsome;
		}
	}
	if ( chanP->channelClosed ) {
		chanP->reading = FALSE;
		chanP->writing = FALSE;
		chanP->accepting = FALSE;
	}
}

/*************************************************************************
function: WriteSomeToChannel - Send some bytes, update counters
Shuts channel down on disconnect.
Does callback notify after writing or detecting channel closed.
If the callback loads more data, this will attempt to send it again.
*/
static
void WriteSomeToChannel (axio_channel_t *chanP)
{
	boolean haveDrained = FALSE;
	int did;

	while ( !chanP->channelClosed && !haveDrained ) {
		did = send(chanP->chan_fd, chanP->writeBuf+chanP->writePos,
					chanP->writeCnt, 0 );
#if AXMS
		if ( -1 == did ) {
			errno = WSAGetLastError();
			if ( errno == WSAEWOULDBLOCK )
				haveDrained = TRUE;
			else if ( errno != WSAEINTR && errno != WSATRY_AGAIN )
				chanP->channelClosed = TRUE;
		}
#else
		if ( -1 == did ) {
			if ( errno == EWOULDBLOCK )
				haveDrained = TRUE;
			else if ( errno != EINTR && errno != EAGAIN )
				chanP->channelClosed = TRUE;
		}
#endif
		else if ( did == 0 ) {	  /* Pipe must be broken */
			chanP->channelClosed = TRUE;
		}
		else if ( did < chanP->writeCnt ) {
			chanP->writePos += did;
			chanP->writeCnt -= did;
			haveDrained = TRUE;	/* reasonable assumption */
		}
		else {
			chanP->writePos = 0;
			chanP->writeCnt = 0;
			haveDrained = TRUE;
		}
	}
	haveDrained = (chanP->writeCnt == 0);
	if ( chanP->writeNtfy )
		(*chanP->writeNtfy)(chanP, (char *)chanP->calledInfo, chanP->writeCnt);
	if ( haveDrained && !chanP->channelClosed && chanP->writeCnt )
		WriteSomeToChannel(chanP);
	if ( chanP->closeOnWrite && !chanP->writeCnt )
		chanP->channelClosed = TRUE;
	if ( chanP->channelClosed ) {
		chanP->reading = FALSE;
		chanP->writing = FALSE;
		chanP->accepting = FALSE;
	}
}

/*************************************************************************
function: AxIOAddToWriteBuffer - Convenience routine
*/
void AxIOAddToWriteBuffer (axio_channel_t *chanP,
		char *buf,
		int cnt,
		int closeOnWrite)	/* if true, close connection when write completes */
{
	int avail,newSize;

	if ( !vfyLiveChannel(chanP) )
		return;

	chanP->closeOnWrite = closeOnWrite;
	if ( !cnt ) {
		if ( chanP->closeOnWrite && !chanP->writeCnt ) {
			chanP->channelClosed = TRUE;
			chanP->reading = FALSE;
			chanP->writing = FALSE;
			chanP->accepting = FALSE;
		}
		return;
	}
	if ( !chanP->writeBuf ) {
		chanP->writeBufSize = MAX(cnt,BASIC_BUF);
		chanP->writeBuf = (char*)malloc(chanP->writeBufSize);
		chanP->writeCnt = 0;
	}
	avail = chanP->writeBufSize - (chanP->writePos + chanP->writeCnt);
	if ( avail < cnt ) {
		newSize = 2*(chanP->writeBufSize+cnt);
		chanP->writeBuf = (char*)realloc(chanP->writeBuf,newSize);
		chanP->writeBufSize = newSize;		
	}
	memcpy(chanP->writeBuf+chanP->writePos+chanP->writeCnt,buf,cnt);
	chanP->writeCnt += cnt;
}

/*************************************************************************
function: AxIOMakeChannelNonBlocking - Read only as much as is there
*/
int AxIOMakeChannelNonBlocking (channel)
int channel;
{
#if AX_X11   /* {  */
#if SUN4 && !SUN4S  /* Posix style nonblock does not seem to work for SunOS */
    return(0 == fcntl(channel,F_SETFL,O_NDELAY));   /* bsd4.2 non-blocking only! */
#else
    return(0 == fcntl(channel, F_SETFL, (fcntl(channel,F_GETFL,0) | O_NONBLOCK)));
#endif
#else     /*  }  {  */
     /* MS-Windows,  NT  */
    unsigned long flag;

    flag = 1;
    ioctlsocket( channel, FIONBIO, &flag );    /* make it non-blocking */

    return( TRUE );
#endif    /*  }  AX_X11 / AXMS  */
}





/*************************************************************************
function: AxIOMakeChannelBlocking - pend until the data is there
*/
int AxIOMakeChannelBlocking (channel)
int channel;
{
#if AX_X11   /* {  */
#if SUN4 && !SUN4S  /* Posix style nonblock does not seem to work for SunOS */
    return(0 == fcntl(channel,F_SETFL,O_NDELAY));   /* bsd4.2 non-blocking only! */
#else
    return(0 == fcntl(channel, F_SETFL, (fcntl(channel,F_GETFL,0) & ~O_NONBLOCK)));
#endif
#else     /*  }  {  */
     /* MS-Windows,  NT  */
    unsigned long flag;

    flag = 0;
    ioctlsocket( channel, FIONBIO, &flag );    /* make it blocking */

    return( TRUE );
#endif    /*  }  AX_X11 / AXMS  */
}






/***********************************************************************/
/* Force data on a channel out the door                                */
/***********************************************************************/
void FlushChannel( axio_channel_t *chanP )
{
    WriteSomeToChannel( chanP );
}


    
