/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: parsmks.c
   Applix ASC version: 5
   Date: Mon Nov 16 13:58:00 1998
*/
#include <utloc.h>
#include <getstr0.h>

#define FS_START	0
#define FS_NUMBER	1
#define FS_DASH		2
#define FS_ERROR	3
#define FS_DONE		4
#define FS_COUNT	(FS_DONE+1)

#define IN_SPACE	0
#define IN_COMMA	1
#define IN_DASH		2
#define IN_NUMBER	3
#define IN_DONE		4
#define IN_COUNT	(IN_DONE+1)

#define ACTERROR	0
#define ACTDONE		1
#define ACTNUM1		2
#define ACTNUM2		3
#define ACTNUM3		4
#define ACTNUM4		5
#define ACTDASH		6
#define ACTNONE		7
#define ACTNUM5		8


#define DIGITS		"0123456789"

struct range {
	long low;
	long high;
};

/*
.BEGIN MANPAGE
.NAME ParseMarks - Interpret a "which numbers?" response.
.DESCRIPTION
	Given a string specifying values for a mark list, ParseMarks() will
	construct a table of range entries by parsing the string for valid range
	specifiers.
.ARGUMENT s
	The address of the string specifying values for a mark list.
.ARGUMENT rngTable
	The address of an array of structures of 2 longs in which to return
	the low bound and high bound of the range indicated in s.
.ARGUMENT rngLength
	The number of entries in the array specified by rngTable.
.RETURN VALUE
	-1 indicating that the range table is too small for the number of
		entered ranges.

 	0 indicating that there was an error in parsing the ranges.

 	> 0 indicating the number of entries in rngTable that are
		filled in.
.NOTES
	The maximum record number that can be specified is 2,130,706,432
	(0x7f000000).

	A range specifier of "all" will return a range of 1 - 2,130,706,432.

	"All" must be the only specifier on the line, i.e.,
	"1-5, all" is not valid and not very useful.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
ParseMarks(s,rngTable,rngLength)
char *s;
struct range *rngTable;
int rngLength;
/* END SYNOPSIS */
{
	static char allstr[STRNORMAL] = "";
	static int allstrlen = -1;
	static int fsmtab[FS_COUNT][IN_COUNT] = {
		ACTNONE,	ACTNONE,	ACTERROR,	ACTNUM1,	ACTDONE,
		ACTNONE,	ACTNUM2,	ACTDASH,	ACTNUM3,	ACTDONE,
		ACTNONE,	ACTNUM4,	ACTERROR,	ACTNUM4,	ACTERROR,
		ACTERROR,	ACTERROR,	ACTERROR,	ACTERROR,	ACTERROR,
		ACTERROR,	ACTERROR,	ACTERROR,	ACTERROR,	ACTERROR
	};
	int state = FS_START,inval,haveNum = FALSE;
	char numtxt[STRSHORT];
	long num;
	int len;
	int rangeCount = 0;

	if(allstrlen == -1) {
		getStr(S0ALL,allstr);
		allstrlen = strlen(allstr);
	}

	while(*s == ' ')
		s++;

	if(!Astrncmp(s,allstr,allstrlen)) {
		s += allstrlen;
		while(*s == ' ')
			s++;
		if(*s)
			return(0);
		rngTable->low = 1;
		rngTable->high = 0x7f000000;
		return(1);
	}

	while(state != FS_DONE && rangeCount < rngLength) {
		switch(*s) {
		  case ' ':
			inval = IN_SPACE;
			break;

		  case ',':
			inval = IN_COMMA;
			break;

		  case '-':
			inval = IN_DASH;
			break;

		  case '0':
		  case '1':
		  case '2':
		  case '3':
		  case '4':
		  case '5':
		  case '6':
		  case '7':
		  case '8':
		  case '9':
			inval = IN_NUMBER;
			if((len = strspn(s,DIGITS)) >= STRSHORT)
				return(0);
			memcpy(numtxt,s,len);
			numtxt[len] = '\0';
			num = (long) atol(numtxt);
			if(num <= 0 || num > 0x7f000000)
				return(0);
			s += len - 1;
			break;

		  case '\0':
			inval = IN_DONE;
			break;

		  default:
			return(0);
		}

		s++;

		switch(fsmtab[state][inval]) {

		  case ACTDONE:
			state = FS_DONE;
			break;

		  case ACTERROR:
			return(0);

		  case ACTNONE:
			break;

		  case ACTDASH:
			state = FS_DASH;
			break;

		  case ACTNUM1:
			haveNum = TRUE;
			rngTable->low = num;
			rngTable->high = num;
			state = FS_NUMBER;
			break;

		  case ACTNUM2:
			if(rngTable->low > rngTable->high)
				return(0);
			rangeCount++;
			rngTable++;
			state = FS_START;
			haveNum = FALSE;
			break;

		  case ACTNUM3:
			haveNum = TRUE;
			if(rngTable->low > rngTable->high)
				return(0);
			rangeCount++;
			rngTable++;
			rngTable->low = num;
			rngTable->high = num;
			break;

		case ACTNUM4:
			rngTable->high = num;
			if(rngTable->low > rngTable->high)
				return(0);
			rangeCount++;
			rngTable++;
			state = FS_START;
			haveNum = FALSE;
			break;

		default:
			return(0);
		}

	}

	if(haveNum)
		rangeCount++;
	if(rangeCount >= rngLength)
		rangeCount = -1;
	return(rangeCount);
}

